# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import shutil
import subprocess
from pathlib import Path

import pytest
from flexmock import flexmock
from specfile import Specfile

from packit.actions import ActionName
from packit.api import Config, PackitAPI
from packit.config import parse_loaded_config
from packit.distgit import DistGit
from packit.local_project import LocalProject
from packit.upstream import GitUpstream
from tests.integration.conftest import mock_spec_download_remote_s
from tests.spellbook import TARBALL_NAME


@pytest.fixture()
def github_release_webhook():
    return {
        "repository": {
            "full_name": "brewery/beer",
            "owner": {"login": "brewery"},
            "name": "beer",
            "html_url": "https://github.com/brewery/beer",
        },
        "release": {
            "body": "Changelog content will be here",
            "tag_name": "0.1.0",
            "created_at": "2019-02-28T18:48:27Z",
            "published_at": "2019-02-28T18:51:10Z",
            "draft": False,
            "prerelease": False,
            "name": "Beer 0.1.0 is gooooood",
        },
        "action": "published",
    }


def test_basic_local_update(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    """basic propose-downstream test: mock remote API, use local upstream and dist-git"""
    u, d, api = api_instance
    mock_spec_download_remote_s(d)
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    flexmock(Specfile).should_call("reload").once()

    api.sync_release(dist_git_branch="main", versions=["0.1.0"])

    assert (d / TARBALL_NAME).is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog


def test_fast_forward_merge_local_update(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream_fast_forward_merge_branches,
):
    _, d, api = api_instance
    _, _ = mock_remote_functionality_upstream_fast_forward_merge_branches
    mock_spec_download_remote_s(d)
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    flexmock(Specfile).should_call("reload").times(1)

    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        fast_forward_merge_branches={"f40"},
    )
    assert (d / TARBALL_NAME).is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog


def test_basic_local_update_no_upload_to_lookaside(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    """basic propose-downstream test: mock remote API, use local upstream and dist-git"""
    u, d, api = api_instance
    mock_spec_download_remote_s(d)
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    flexmock(Specfile).should_call("reload").once()
    flexmock(DistGit).should_call("upload_to_lookaside_cache").with_args(
        archives=[d / "beerware-0.1.0.tar.gz"],
        pkg_tool="",
        offline=True,
    ).once()

    api.package_config.upload_sources = False
    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
    )

    assert (d / TARBALL_NAME).is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog


def test_basic_local_update_use_downstream_specfile(
    cwd_upstream,
    api_instance,
    distgit_and_remote,
    mock_remote_functionality_upstream,
    caplog,
):
    # log specfile debug messages
    caplog.set_level(logging.DEBUG)

    u, d, api = api_instance
    # remove the upstream specfile and push the tag that will be checked out
    u.joinpath("beer.spec").unlink()
    subprocess.check_call(
        ["git", "commit", "-m", "remove spec", "-a"],
        cwd=str(u),
    )
    subprocess.check_call(["git", "tag", "0.1.0", "-f"])
    mock_spec_download_remote_s(d)
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()

    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        use_downstream_specfile=True,
    )

    assert (d / TARBALL_NAME).is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog

    # do this second time to see whether the specfile is updated correctly
    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        use_downstream_specfile=True,
    )

    assert (d / TARBALL_NAME).is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog

    assert changelog.count("0.1.0-1") == 1


def test_basic_local_update_use_downstream_specfile_non_git_upstream(
    cwd_upstream,
    api_instance,
    distgit_and_remote,
    mock_remote_functionality_upstream,
    caplog,
):
    # log specfile debug messages
    caplog.set_level(logging.DEBUG)

    u, d, api = api_instance
    api.non_git_upstream = True
    api._up = None

    # remove the upstream specfile and push the tag that will be checked out
    u.joinpath("beer.spec").unlink()
    subprocess.check_call(
        ["git", "commit", "-m", "remove spec", "-a"],
        cwd=str(u),
    )
    subprocess.check_call(["git", "tag", "0.1.0", "-f"])
    mock_spec_download_remote_s(d)
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    flexmock(api.up).should_receive("sync_files")

    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        use_downstream_specfile=True,
    )

    assert (d / TARBALL_NAME).is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog

    # do this second time to see whether the specfile is updated correctly
    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        use_downstream_specfile=True,
    )

    assert (d / TARBALL_NAME).is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog

    assert changelog.count("0.1.0-1") == 1


def test_basic_local_update_with_multiple_sources(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    """basic propose-downstream test: mock remote API, use local upstream and dist-git"""
    u, d, api = api_instance
    mock_spec_download_remote_s(d, files_to_create=["the_source.tar.gz"])
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    for git_path in [u, d]:
        with Specfile(git_path / "beer.spec", autosave=True).sources() as sources:
            sources.append("https://the.second.source/the_source.tar.gz")
        subprocess.check_call(["git", "add", "beer.spec"], cwd=git_path)
        subprocess.check_call(
            ["git", "commit", "-m", "Added new source to specfile"],
            cwd=git_path,
        )
    subprocess.check_call(["git", "tag", "0.1.0", "-f"], cwd=u)

    dist_git_first_source = d / TARBALL_NAME
    dist_git_second_source = d / "the_source.tar.gz"
    flexmock(api.dg).should_call("upload_to_lookaside_cache").with_args(
        archives=[dist_git_first_source, dist_git_second_source],
        pkg_tool="",
        offline=False,
    )

    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
    )

    assert dist_git_first_source.is_file()
    assert dist_git_second_source.is_file()

    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    with spec.sources() as sources:
        newly_added_source = sources[1]
        assert (
            newly_added_source.expanded_location
            == "https://the.second.source/the_source.tar.gz"
        )
        assert newly_added_source.filename == "the_source.tar.gz"
    assert spec.sources()
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog


def test_basic_local_update_with_adding_second_source(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    """basic propose-downstream test: mock remote API, use local upstream and dist-git"""
    u, d, api = api_instance
    mock_spec_download_remote_s(d, files_to_create=["the_source.tar.gz"])
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    with Specfile(u / "beer.spec", autosave=True).sources() as sources:
        sources.append("https://the.second.source/the_source.tar.gz")
    subprocess.check_call(["git", "add", "beer.spec"], cwd=u)
    subprocess.check_call(
        ["git", "commit", "-m", "Added new source to specfile"],
        cwd=u,
    )
    subprocess.check_call(["git", "tag", "0.1.0", "-f"], cwd=u)

    dist_git_first_source = d / TARBALL_NAME
    dist_git_second_source = d / "the_source.tar.gz"
    flexmock(api.dg).should_call("upload_to_lookaside_cache").with_args(
        archives=[dist_git_first_source, dist_git_second_source],
        pkg_tool="",
        offline=False,
    )

    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
    )

    assert dist_git_first_source.is_file()
    assert dist_git_second_source.is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    with spec.sources() as sources:
        newly_added_source = sources[1]
        assert (
            newly_added_source.expanded_location
            == "https://the.second.source/the_source.tar.gz"
        )
        assert newly_added_source.filename == "the_source.tar.gz"
    assert spec.sources()
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog


def test_basic_local_update_with_adding_second_local_source(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    """basic propose-downstream test: mock remote API, use local upstream and dist-git"""
    u, d, api = api_instance
    mock_spec_download_remote_s(d, files_to_create=["the_source.tar.gz"])
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    with Specfile(u / "beer.spec", autosave=True).sources() as sources:
        sources.append("the_source.tar.gz")
    subprocess.check_call(["git", "add", "beer.spec"], cwd=u)
    subprocess.check_call(
        ["git", "commit", "-m", "Added new source to specfile"],
        cwd=u,
    )
    subprocess.check_call(["git", "tag", "0.1.0", "-f"], cwd=u)

    dist_git_first_source = d / TARBALL_NAME
    dist_git_second_source = d / "the_source.tar.gz"
    flexmock(api.dg).should_call("upload_to_lookaside_cache").with_args(
        archives=[dist_git_first_source, dist_git_second_source],
        pkg_tool="",
        offline=False,
    )

    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
    )

    assert dist_git_first_source.is_file()
    assert dist_git_second_source.is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    with spec.sources() as sources:
        newly_added_source = sources[1]
        assert newly_added_source.expanded_location == "the_source.tar.gz"
        assert newly_added_source.filename == "the_source.tar.gz"
    assert spec.sources()
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog


def test_basic_local_update_with_adding_second_local_source_tracked_by_git(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    """basic propose-downstream test: mock remote API, use local upstream and dist-git"""
    u, d, api = api_instance
    mock_spec_download_remote_s(d, files_to_create=["the_source.tar.gz"])
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    with Specfile(u / "beer.spec", autosave=True).sources() as sources:
        sources.append("README")
    subprocess.check_call(["git", "add", "beer.spec"], cwd=u)
    subprocess.check_call(
        ["git", "commit", "-m", "Added new source to specfile"],
        cwd=u,
    )
    subprocess.check_call(["git", "tag", "0.1.0", "-f"], cwd=u)

    dist_git_first_source = d / TARBALL_NAME
    dist_git_second_source = d / "README"

    flexmock(api.dg).should_call("upload_to_lookaside_cache").with_args(
        archives=[dist_git_first_source],
        pkg_tool="",
        offline=False,
    )

    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
    )

    assert dist_git_first_source.is_file()
    assert dist_git_second_source.is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert spec.sources()
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog


def test_basic_local_update_with_removing_second_source(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    """basic propose-downstream test: mock remote API, use local upstream and dist-git"""
    u, d, api = api_instance
    mock_spec_download_remote_s(d, files_to_create=[])
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    with Specfile(d / "beer.spec", autosave=True).sources() as sources:
        sources.append("https://the.second.source/the_source.tar.gz")
    subprocess.check_call(["git", "add", "beer.spec"], cwd=d)
    subprocess.check_call(
        ["git", "commit", "-m", "Added new source to specfile"],
        cwd=d,
    )
    api.dg.specfile.reload()

    dist_git_first_source = d / TARBALL_NAME
    dist_git_second_source = d / "the_source.tar.gz"
    flexmock(api.dg).should_call("upload_to_lookaside_cache").with_args(
        archives=[dist_git_first_source],
        pkg_tool="",
        offline=False,
    )

    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
    )

    assert dist_git_first_source.is_file()
    assert not dist_git_second_source.is_file()

    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    with spec.sources() as sources:
        assert len(sources) == 1  # The second source should be removed
    assert spec.sources()
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog


def test_local_update_generated_spec(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    """Check that specfile can be generated on clone."""
    u, d, api = api_instance
    mock_spec_download_remote_s(d)
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    flexmock(GitUpstream).should_receive("get_latest_released_version").and_return(
        "0.1.0",
    )

    # Simulate generation by moving the spec to a different location
    # We are checking two things:
    # * spec-file is not used before the post-upstream-clone
    # * the version is get from the release state
    current_spec_location = u / "beer.spec"
    new_spec_location = u / ".." / "tmp.spec"
    shutil.move(current_spec_location, new_spec_location)
    subprocess.check_call(["git", "add", "beer.spec"], cwd=u)
    subprocess.check_call(["git", "commit", "-m", "Spec removed from upstream"], cwd=u)
    api.up.package_config.actions = {
        ActionName.post_upstream_clone: [
            f"mv {new_spec_location} {current_spec_location}",
        ],
    }

    api.sync_release(dist_git_branch="main")

    assert (d / TARBALL_NAME).is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog


def test_basic_local_update_reset_after_exception(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    """check whether the distgit repo is not dirty after exception is raised"""
    u, d, api = api_instance
    mock_spec_download_remote_s(d)
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    flexmock(api).should_receive("_handle_sources").and_raise(Exception)
    with pytest.raises(Exception):
        api.sync_release("main", "0.1.0")

    assert not api.dg.is_dirty()


def test_basic_local_update_copy_upstream_release_description(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    """basic propose-downstream test: mock remote API, use local upstream and dist-git,
    set copy_upstream_release_description in package config to True"""
    u, d, api = api_instance
    mock_spec_download_remote_s(d)
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    release = flexmock(body="Some description of the upstream release", url="some-url")
    api.up.local_project.git_project = flexmock(
        get_release=lambda name, tag_name: release,
    )
    api.package_config.copy_upstream_release_description = True
    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
    )

    assert (d / TARBALL_NAME).is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)

    assert (
        """- 0.1.0-1
Some description of the upstream release
"""
        in changelog
    )

    assert "0.0.0" in changelog
    assert "0.1.0" in changelog


def test_basic_local_update_using_distgit(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    """basic propose-downstream test: mock remote API, use local upstream and dist-git"""
    u, d, api = api_instance
    mock_spec_download_remote_s(d)

    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
    )

    assert (d / TARBALL_NAME).is_file()
    spec = Specfile(d / "beer.spec")
    assert spec.expanded_version == "0.1.0"

    with spec.sections() as sections:
        package_section = sections.package

    assert package_section[2] == "# some change"
    assert package_section[4] == "Name:           beer"
    assert package_section[5] == "Version:        0.1.0"
    assert package_section[6] == "Release:        1%{?dist}"
    assert package_section[7] == "Summary:        A tool to make you happy"

    assert (d / "README.packit").is_file()
    # assert that we have changelog entries for both versions
    with spec.sections() as sections:
        changelog = "\n".join(sections.changelog)
    assert "0.0.0" in changelog
    assert "0.1.0" in changelog


def test_basic_local_update_direct_push(
    cwd_upstream,
    api_instance,
    distgit_and_remote,
    mock_remote_functionality_upstream,
):
    """basic propose-downstream test: mock remote API, use local upstream and dist-git"""
    u, d, api = api_instance
    _, distgit_remote = distgit_and_remote
    mock_spec_download_remote_s(d)

    api.sync_release(dist_git_branch="main", versions=["0.1.0"], create_pr=False)

    remote_dir_clone = Path(f"{distgit_remote}-clone")
    subprocess.check_call(
        ["git", "clone", distgit_remote, str(remote_dir_clone)],
        cwd=str(remote_dir_clone.parent),
    )

    spec = Specfile(remote_dir_clone / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (remote_dir_clone / "README.packit").is_file()


def test_update_downstream_changelog_even_if_has_autochangelog(
    cwd_upstream,
    api_instance,
    distgit_with_autochangelog_and_remote,
    mock_remote_functionality_downstream_autochangelog,
):
    """Check that a new entry is added to the %changelog section of the the spec-file in dist-git,
    when sync_changelog is set, even if the spec-file in dist-git uses %autochangelog"""
    u, d, api = api_instance
    _, distgit_remote = distgit_with_autochangelog_and_remote

    api.package_config.sync_changelog = True
    api.sync_release(
        dist_git_branch="main",
        versions=["0.1.0"],
        create_pr=False,
        add_new_sources=False,
    )

    assert api.dg.specfile.version == "0.1.0"
    with api.dg.specfile.sections() as sections:
        assert (
            "* Mon Feb 25 2019 Tomas Tomecek <ttomecek@redhat.com> - 0.1.0-1"
            in sections.changelog
        )


def test_basic_local_update_direct_push_no_dg_spec(
    cwd_upstream,
    api_instance,
    distgit_and_remote,
    mock_remote_functionality_upstream,
):
    u, d, api = api_instance
    d.joinpath("beer.spec").unlink()
    subprocess.check_call(
        ["git", "commit", "-m", "remove spec", "-a"],
        cwd=str(d),
    )
    _, distgit_remote = distgit_and_remote
    mock_spec_download_remote_s(d)

    api.sync_release(dist_git_branch="main", versions=["0.1.0"], create_pr=False)

    remote_dir_clone = Path(f"{distgit_remote}-clone")
    subprocess.check_call(
        ["git", "clone", distgit_remote, str(remote_dir_clone)],
        cwd=str(remote_dir_clone.parent),
    )

    spec = Specfile(remote_dir_clone / "beer.spec")
    assert spec.expanded_version == "0.1.0"
    assert (remote_dir_clone / "README.packit").is_file()


def test_basic_local_update_from_downstream(
    cwd_upstream,
    api_instance,
    mock_remote_functionality_upstream,
):
    flexmock(LocalProject, _parse_namespace_from_git_url=lambda: None)
    u, d, api = api_instance

    api.sync_from_downstream("main", "main", True)

    new_upstream = api.up.local_project.working_dir
    assert (new_upstream / "beer.spec").is_file()
    spec = Specfile(new_upstream / "beer.spec")
    assert spec.expanded_version == "0.0.0"


def test_local_update_with_specified_tag_template():
    c = Config()
    pc = parse_loaded_config(
        {
            "specfile_path": "beer.spec",
            "synced_files": ["beer.spec"],
            "upstream_package_name": "beerware",
            "downstream_package_name": "beer",
            "upstream_tag_template": "v{version}",
            "create_pr": False,
        },
    )
    api = PackitAPI(c, pc)

    assert (
        api.up.package_config.upstream_tag_template.format(version="0.1.0") == "v0.1.0"
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock

from packit.utils.upstream_version import get_upstream_version, requests


@pytest.mark.parametrize(
    "package, version, exception",
    [
        ("libtiff", "4.4.0", None),
        ("tiff", "4.4.0", None),
        ("python-specfile", "0.5.0", None),
        ("specfile", "0.5.0", None),
        ("python3-specfile", None, None),
        ("mock", "3.1-1", None),
        ("packitos", "0.56.0", None),
        ("packitos", "0.56.0", requests.exceptions.SSLError),
        ("packit", None, None),
    ],
)
def test_get_upstream_version(package, version, exception):
    def mocked_get(url, params, *_, **__):
        if exception is not None:
            raise exception()
        packages = {
            "libtiff": "tiff",
            "python-specfile": "specfile",
            "mock": "mock",
        }
        projects = {
            "tiff": "4.4.0",
            "specfile": "0.5.0",
            "mock": "3.1-1",
            "packitos": "0.56.0",
        }
        if url.endswith("projects"):
            project, version = next(
                iter(
                    (k, v)
                    for k, v in projects.items()
                    if k.startswith(params["pattern"])
                ),
                (None, None),
            )
            items = [{"name": project, "version": version}] if project else []
            return flexmock(ok=True, json=lambda: {"projects": items})
        package_name = url.split("/")[-1]
        project = packages.get(package_name)
        version = projects.get(project)
        if version:
            return flexmock(ok=True, json=lambda: {"version": version})
        return flexmock(ok=False)

    flexmock(requests).should_receive("get").replace_with(mocked_get)

    if exception is None:
        assert get_upstream_version(package) == version
    else:
        assert get_upstream_version(package) is None



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from typing import Optional

import pytest

from packit.config.common_package_config import _construct_dist_git_instance
from packit.constants import DISTGIT_INSTANCES
from packit.dist_git_instance import DistGitInstance


@pytest.mark.parametrize(
    "base_url, namespace, pkg_tool, sig, expected_dg_instance",
    (
        (None, None, "fedpkg", None, DISTGIT_INSTANCES["fedpkg"]),
        (None, None, "centpkg", None, DISTGIT_INSTANCES["centpkg"]),
        (
            None,
            None,
            "centpkg-sig",
            "cloud",
            DistGitInstance(
                hostname="gitlab.com",
                alternative_hostname=None,
                namespace="CentOS/cloud/rpms",
            ),
        ),
    ),
)
def test_construct_dg_instance(
    base_url: str,
    namespace: str,
    pkg_tool: str,
    sig: Optional[str],
    expected_dg_instance: DistGitInstance,
):
    assert (
        _construct_dist_git_instance(
            base_url=base_url,
            namespace=namespace,
            pkg_tool=pkg_tool,
            sig=sig,
        )
        == expected_dg_instance
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
from importlib.metadata import version

import click

from packit.cli.build import build
from packit.cli.create_update import create_update
from packit.cli.dist_git import dist_git
from packit.cli.init import init
from packit.cli.prepare_sources import prepare_sources
from packit.cli.propose_downstream import propose_downstream, pull_from_upstream
from packit.cli.push_updates import push_updates
from packit.cli.scan_in_osh import scan_in_osh
from packit.cli.source_git import source_git
from packit.cli.srpm import srpm
from packit.cli.status import status
from packit.cli.sync_from_downstream import sync_from_downstream
from packit.cli.validate_config import validate_config
from packit.config import Config, get_context_settings
from packit.utils.logging import set_logging

logger = logging.getLogger("packit")


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        if cmd_name == "propose-update":
            click.secho(
                "WARNING: 'packit propose-update' is deprecated and will be removed. "
                "Use 'packit propose-downstream' instead.",
                fg="yellow",
            )
            return click.Group.get_command(self, ctx, "propose-downstream")
        return click.Group.get_command(self, ctx, cmd_name)


@click.group("packit", cls=AliasedGroup, context_settings=get_context_settings())
@click.option("-d", "--debug", is_flag=True, help="Enable debug logs.")
@click.option("--fas-user", help="Fedora Account System username.")
@click.option("-k", "--keytab", help="Path to FAS keytab file.")
@click.option(
    "--remote",
    default=None,
    help=(
        "Name of the remote to discover upstream project URL, "
        "If this is not specified, default to the first remote."
    ),
)
@click.option(
    "-c",
    "--config",
    "package_config_path",
    help="Path to package configuration file (defaults to .packit.yaml or packit.yaml)",
)
@click.version_option(version=version("packitos"), message="%(version)s")
@click.pass_context
def packit_base(ctx, debug, fas_user, keytab, remote, package_config_path):
    """Integrate upstream open source projects into Fedora operating system."""
    if debug:
        # to be able to logger.debug() also in get_user_config()
        set_logging(level=logging.DEBUG)

    c = Config.get_user_config()
    c.debug = debug or c.debug
    c.fas_user = fas_user or c.fas_user
    c.keytab_path = keytab or c.keytab_path
    c.upstream_git_remote = remote or c.upstream_git_remote
    c.package_config_path = package_config_path or c.package_config_path
    ctx.obj = c

    if ctx.obj.debug:
        set_logging(level=logging.DEBUG)
        set_logging(logger_name="sandcastle", level=logging.DEBUG)
    else:
        set_logging(level=logging.INFO)

    packit_version = version("packitos")
    logger.debug(f"Packit {packit_version} is being used.")


packit_base.add_command(propose_downstream)
packit_base.add_command(pull_from_upstream)
packit_base.add_command(sync_from_downstream)
packit_base.add_command(build)
packit_base.add_command(create_update)
packit_base.add_command(push_updates)
packit_base.add_command(srpm)
packit_base.add_command(status)
packit_base.add_command(init)
packit_base.add_command(validate_config)
packit_base.add_command(source_git)
packit_base.add_command(prepare_sources)
packit_base.add_command(dist_git)
packit_base.add_command(scan_in_osh)

if __name__ == "__main__":
    packit_base()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
"""
Initialise the Bodhi client.

Bodhi 6+ prints a URL which you need to open in browser and paste 'code=XXX...' into
terminal which the bodhi-client will save to `~/.config/bodhi/client.json`.
Bodhi 6 does not use username, password nor keytab.
"""
import os

from bodhi.client.bindings import BodhiClient


def get_bodhi_client() -> BodhiClient:
    """
    Provide an instance of BodhiClient

    Note for tests:
    * Don't mock the Bodhi instantiation via
      `flexmock(bodhi.client.bindings.BodhiClient).new_instances(bodhi_instance_mock)`
      what can affect other tests.
    * Mock this function instead -- for example:
      `flexmock(aliases).should_receive("get_bodhi_client").and_return(bodhi_instance_mock)`
    """
    return BodhiClient(
        oidc_storage_path=os.path.join(os.environ["HOME"], "bodhi-client.json"),
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import shutil
from pathlib import Path

import pytest
from specfile import Specfile

from tests.spellbook import SPECFILE


@pytest.fixture(scope="function")
def specfile(tmp_path):
    specfile_path = tmp_path / "beer.spec"
    shutil.copyfile(SPECFILE, specfile_path)
    return specfile_path


def test_write_spec_content(specfile):
    data = "new line 1\n"
    spec = Specfile(specfile, autosave=True)
    with spec.sections() as sections:
        sections.description = [data]

    assert "new line 1" in specfile.read_text()


simple_patch = """\
# A bright yellow patch
Patch1: yellow.patch
"""

patch_with_url = """\
# A bright yellow patch
Patch0001  :  https://inter.netz/yellow.patch\n
"""

multiple_patches = """\
# An interesting purple
Patch0200: purple.patch

# Orange is a fruit
# And orange is a colour
Patch0301: orange.patch
"""

hanging_comments = """\
# This comment doesn't belong to any patch

Patch: surprise.patch
"""

no_space = """\
# An interesting purple
Patch0200: purple.patch
# Orange is a fruit
Patch0301: orange.patch
"""


@pytest.mark.parametrize(
    "lines,files,expectation",
    [
        (simple_patch, ["yellow.patch"], {"yellow.patch": ["A bright yellow patch"]}),
        (patch_with_url, ["yellow.patch"], {"yellow.patch": ["A bright yellow patch"]}),
        (
            multiple_patches,
            ["purple.patch", "orange.patch"],
            {
                "purple.patch": ["An interesting purple"],
                "orange.patch": ["Orange is a fruit", "And orange is a colour"],
            },
        ),
        (hanging_comments, ["surprise.patch"], {"surprise.patch": []}),
        (
            no_space,
            ["purple.patch", "orange.patch"],
            {
                "purple.patch": ["An interesting purple"],
                "orange.patch": ["Orange is a fruit"],
            },
        ),
    ],
    ids=[
        "simple-patch",
        "patch-with-url",
        "multiple-patches",
        "hanging-comments",
        "no-space",
    ],
)
def test_read_patch_comments(specfile, lines, files, expectation):
    """Check reading comment lines that belong to patches"""
    content = specfile.read_text()
    content = content.replace("### patches ###\n", lines)
    specfile.write_text(content)
    for patch_file in files:
        Path(specfile.parent, patch_file).touch()
    spec = Specfile(specfile, sourcedir=specfile.parent, autosave=True)
    with spec.patches() as patches:
        comments = {p.filename: [c.text for c in p.comments] for p in patches}
        assert comments == expectation


@pytest.mark.parametrize(
    "lines,digits",
    [
        ("Patch0001 : some.patch\n", 4),
        ("Patch003000 : some.patch\n", 6),
        ("Patch: some.patch\n", 0),
        ("Patch21: some.patch\n", 1),
        ("Patch9: some.patch\n", 1),
    ],
)
def test_patch_id_digits(specfile, lines, digits):
    """Check detecting the number of digits used for patch IDs (indices)"""
    content = specfile.read_text()
    content = content.replace("### patches ###\n", lines)
    specfile.write_text(content)
    spec = Specfile(specfile, sourcedir=specfile.parent, autosave=True)
    with spec.patches() as patches:
        assert patches[0].number_digits == digits


def test_remove_patches(specfile):
    """Check patches being removed from a specfile"""
    no_patches = specfile.read_text().replace("\n### patches ###\n", "")
    patches = specfile.read_text().replace(
        "### patches ###\n",
        """\
# Some comment line to be removed
Patch1: yellow.patch

Patch2: blue.patch
# One
# Or more lines
Patch : dark.patch
""",
    )
    specfile.write_text(patches)
    spec = Specfile(specfile, sourcedir=specfile.parent, autosave=True)
    with spec.patches() as patches:
        patches.clear()
    assert specfile.read_text() == no_patches


def test_remove_patches_no_blanklines(specfile):
    no_blanks = specfile.read_text().replace("\n\n", "\n")
    no_patches = no_blanks.replace("\n### patches ###\n", "\n")
    patches = no_blanks.replace(
        "### patches ###\n",
        """\
# Some comment line to be removed
Patch1: yellow.patch
Patch2: blue.patch
# One
# Or more lines
Patch : dark.patch
""",
    )
    specfile.write_text(patches)
    spec = Specfile(specfile, sourcedir=specfile.parent, autosave=True)
    with spec.patches() as patches:
        patches.clear()
    assert specfile.read_text() == no_patches



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import pytest
from requre.cassette import DataTypes
from requre.modules_decorate_all_methods import record_requests_module

from packit.exceptions import ImageBuilderError
from packit.vm_image_build import ImageBuilder
from tests_recording.testbase import PackitTest


@record_requests_module
class TestLocalProject(PackitTest):
    """Test Image Builder API and Red Hat Auth requests"""

    def cassette_setup(self, cassette):
        """requre requires this method to be present"""
        cassette.data_miner.data_type = DataTypes.Dict

    def test_get_token(self):
        """Test PR checkout with and without merging"""

        ib = ImageBuilder(refresh_token=self.config.redhat_api_refresh_token)
        ib.refresh_auth()

        assert ib.session.headers["Authorization"] == f"Bearer {ib._access_token}"

    def test_token_auto_refresh(self):
        """make sure that 401 requests obtain a new API token"""

        ib = ImageBuilder(refresh_token=self.config.redhat_api_refresh_token)
        # when regenerating this, go to https://console.redhat.com/insights/image-builder
        # and find a successful build of an image
        response_json = ib.image_builder_request(
            "GET",
            "composes/e31d475a-1d32-4cac-844e-a6a613f80439",
        ).json()
        assert response_json["image_status"]["status"] == "success"

    def test_bad_request(self):
        ib = ImageBuilder(refresh_token=self.config.redhat_api_refresh_token)
        with pytest.raises(ImageBuilderError):
            ib.create_image("fedora-rawhide", "foo-bar", {}, {}, "")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from packit.utils.versions import compare_versions


@pytest.mark.parametrize(
    "a, b, result",
    [
        ("1.0", "1.0", 0),
        ("1.0", "2.0", -1),
        ("2.0", "1.0", 1),
        ("invalid", "invalid", 0),
        ("", "invalid", -1),
        ("invalid", "0.0", -1),
        ("0.0", "", 1),
    ],
)
def test_compare_versions(a, b, result):
    if result == 0:
        assert compare_versions(a, b) == result
    else:
        assert compare_versions(a, b) / result > 0



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from packit.constants import DISTRO_DIR, FROM_DIST_GIT_TOKEN
from packit.exceptions import PackitException


def test_update_source_git_sources_changed(
    sourcegit_and_remote,
    distgit_and_remote,
    api_instance_update_source_git,
):
    """Checks that an error is thrown when sources in dist-git were modified."""
    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    (distgit / "sources").write_text("abcd")
    api_instance_update_source_git.dg.commit("Update sources", "")
    with pytest.raises(PackitException):
        api_instance_update_source_git.update_source_git("HEAD~1..")


def test_update_source_git_patch_changed(
    sourcegit_and_remote,
    distgit_and_remote,
    api_instance_update_source_git,
):
    """Checks that an error is thrown when a patch was modified in the
    given commit."""
    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote

    patch = "something.patch"
    (distgit / patch).write_text("abcd")
    api_instance_update_source_git.dg.specfile.add_patch(patch)
    api_instance_update_source_git.dg.commit("Add a patch", "")
    with pytest.raises(PackitException):
        api_instance_update_source_git.update_source_git("HEAD~1..")


def test_update_source_git_gitignore_empty_commit(
    sourcegit_and_remote,
    distgit_and_remote,
    api_instance_update_source_git,
):
    """Checks that inapplicable gitignore changes do not cause an error
    when a commit contains only the gitignore change."""
    sourcegit, _ = sourcegit_and_remote
    (sourcegit / DISTRO_DIR / ".gitignore").write_text("# Reset gitignore\n!*\n")
    api_instance_update_source_git.up.commit(
        "Reset gitignore",
        "",
        trailers=[
            (
                FROM_DIST_GIT_TOKEN,
                api_instance_update_source_git.dg.local_project.git_repo.head.commit.hexsha,
            ),
        ],
    )

    distgit, _ = distgit_and_remote
    (distgit / ".gitignore").write_text("# Former gitignore\na\nb\n")
    api_instance_update_source_git.dg.commit("Setup gitignore", "")

    # Should not raise on gitignore changes
    api_instance_update_source_git.update_source_git("HEAD~1..")
    # There is nothing to commit in the gitignore patch, head should stay the same
    assert (
        "Reset"
        in api_instance_update_source_git.up.local_project.git_repo.head.commit.message
    )


def test_update_source_git_gitignore(
    sourcegit_and_remote,
    distgit_and_remote,
    api_instance_update_source_git,
):
    """Checks that inapplicable gitignore changes do not cause an error but
    the rest of the commit is applied."""
    sourcegit, _ = sourcegit_and_remote
    (sourcegit / DISTRO_DIR / ".gitignore").write_text("# Reset gitignore\n!*\n")
    api_instance_update_source_git.up.commit(
        "Reset gitignore",
        "",
        trailers=[
            (
                FROM_DIST_GIT_TOKEN,
                api_instance_update_source_git.dg.local_project.git_repo.head.commit.hexsha,
            ),
        ],
    )

    distgit, _ = distgit_and_remote
    (distgit / ".gitignore").write_text("# Former gitignore\na\nb\n")
    content = "abcd"
    (distgit / "c").write_text(content)
    api_instance_update_source_git.dg.commit("Setup gitignore", "")

    # Should not raise on gitignore changes
    api_instance_update_source_git.update_source_git("HEAD~1..")
    assert content == (sourcegit / DISTRO_DIR / "c").read_text()
    assert (
        "Setup"
        in api_instance_update_source_git.up.local_project.git_repo.head.commit.message
    )


@pytest.mark.parametrize(
    "revision_range",
    [
        pytest.param("HEAD~1..", id="revision_range_set"),
        pytest.param(None, id="revision_range_not_set"),
    ],
)
def test_update_source_git(
    sourcegit_and_remote,
    distgit_and_remote,
    api_instance_update_source_git,
    revision_range,
):
    """Updates source-git based on multiple 'extra' commits in the dist-git,
    tests the various possible types of changes of the commit:

    - file creation
    - file modification
    - file rename
    - file removal
    """
    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote

    # Create a new file
    new_file = "test_file.txt"
    content = "abcd"
    (distgit / new_file).write_text(content)
    api_instance_update_source_git.dg.commit("Add", "")
    api_instance_update_source_git.update_source_git(revision_range)
    assert (sourcegit / DISTRO_DIR / new_file).read_text() == content
    assert (
        "Add"
        in api_instance_update_source_git.up.local_project.git_repo.head.commit.message
    )
    assert (
        f"From-dist-git-commit: "
        f"{api_instance_update_source_git.dg.local_project.git_repo.head.commit.hexsha}"
        in api_instance_update_source_git.up.local_project.git_repo.head.commit.message
    )

    # Modify the existing file
    extra_content = "\ndefgh"
    with open(distgit / new_file, "a") as file:
        file.write(extra_content)
    api_instance_update_source_git.dg.commit("Modify", "")
    api_instance_update_source_git.update_source_git(revision_range)
    assert (sourcegit / DISTRO_DIR / new_file).read_text() == content + extra_content
    assert (
        "Modify"
        in api_instance_update_source_git.up.local_project.git_repo.head.commit.message
    )
    assert (
        f"From-dist-git-commit: "
        f"{api_instance_update_source_git.dg.local_project.git_repo.head.commit.hexsha}"
        in api_instance_update_source_git.up.local_project.git_repo.head.commit.message
    )

    # Rename a file
    new_name = "test.txt"
    (distgit / new_file).rename(distgit / new_name)
    api_instance_update_source_git.dg.commit("Rename", "")
    api_instance_update_source_git.update_source_git(revision_range)
    assert not (sourcegit / DISTRO_DIR / new_file).exists()
    with open(sourcegit / DISTRO_DIR / new_name) as file:
        assert file.read() == content + extra_content
    assert (
        "Rename"
        in api_instance_update_source_git.up.local_project.git_repo.head.commit.message
    )
    assert (
        f"From-dist-git-commit: "
        f"{api_instance_update_source_git.dg.local_project.git_repo.head.commit.hexsha}"
        in api_instance_update_source_git.up.local_project.git_repo.head.commit.message
    )

    # Delete a file
    (distgit / new_name).unlink()
    api_instance_update_source_git.dg.commit("Delete", "Some file")
    api_instance_update_source_git.update_source_git(revision_range)
    assert not (sourcegit / DISTRO_DIR / new_name).exists()
    assert (
        """Delete

Some file"""
        in api_instance_update_source_git.up.local_project.git_repo.head.commit.message
    )
    assert (
        f"From-dist-git-commit: "
        f"{api_instance_update_source_git.dg.local_project.git_repo.head.commit.hexsha}"
        in api_instance_update_source_git.up.local_project.git_repo.head.commit.message
    )


def test_update_source_git_diverged(
    sourcegit_and_remote,
    distgit_and_remote,
    api_instance_update_source_git,
):
    """Check that exception is raised when there's
    a commit in source-git which has no origin in dist-git.
    """
    sourcegit, _ = sourcegit_and_remote
    (sourcegit / DISTRO_DIR / "test").write_text("foo")
    api_instance_update_source_git.up.commit("test foor", "")

    distgit, _ = distgit_and_remote
    (distgit / "bar").write_text("bar")
    api_instance_update_source_git.dg.commit("test bar", "")

    with pytest.raises(PackitException) as exc:
        api_instance_update_source_git.update_source_git("HEAD~1..")
    assert " have diverged." in str(exc.value)


def test_update_source_git_source_git_not_pristine(
    sourcegit_and_remote,
    distgit_and_remote,
    api_instance_update_source_git,
):
    """Check that exception is raised when the
    source-git repo is not pristine.
    """
    sourcegit, _ = sourcegit_and_remote
    (sourcegit / "new_file").write_text("abcd")

    distgit, _ = distgit_and_remote
    (distgit / "bar").write_text("bar")
    api_instance_update_source_git.dg.commit("test bar", "")

    with pytest.raises(PackitException) as exc:
        api_instance_update_source_git.update_source_git("HEAD~..")
    assert "is not pristine" in str(exc.value)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Display status
"""

import logging
import os

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api, iterate_packages
from packit.config import get_context_settings, pass_config
from packit.constants import (
    PACKAGE_LONG_OPTION,
    PACKAGE_OPTION_HELP,
    PACKAGE_SHORT_OPTION,
)

logger = logging.getLogger(__name__)


@click.command("status", context_settings=get_context_settings())
@click.option(
    PACKAGE_SHORT_OPTION,
    PACKAGE_LONG_OPTION,
    multiple=True,
    help=PACKAGE_OPTION_HELP.format(action="update"),
)
@click.argument("path_or_url", type=LocalProjectParameter(), default=os.path.curdir)
@pass_config
@cover_packit_exception
@iterate_packages
def status(config, path_or_url, package_config):
    """
    Display status.

    \b
    - latest downstream pull requests
    - versions from all downstream branches
    - latest upstream releases
    - latest builds in Koji
    - latest builds in Copr
    - latest updates in Bodhi
    """

    api = get_packit_api(
        config=config,
        package_config=package_config,
        local_project=path_or_url,
    )
    api.status()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import subprocess
from logging import getLogger

import pytest
import rpm
from flexmock import flexmock

from packit.actions import ActionName
from packit.config.package_config import PackageConfig
from packit.distgit import DistGit
from packit.utils.changelog_helper import ChangelogHelper

logger = getLogger(__name__)


@pytest.fixture
def package_config():
    yield PackageConfig()


@pytest.fixture
def upstream(upstream_instance):
    _, ups = upstream_instance
    yield ups


@pytest.fixture
def downstream():
    yield flexmock(DistGit)


def test_srpm_action(upstream, downstream):
    package_config = upstream.package_config
    package_config.actions = {
        ActionName.changelog_entry: [
            "echo - hello from test_srpm_action   ",
        ],
    }

    ChangelogHelper(upstream, downstream, package_config).prepare_upstream_locally(
        "0.1.0",
        "abc123a",
        True,
        None,
    )
    with upstream.specfile.sections() as sections:
        assert "- hello from test_srpm_action" in sections.changelog


def test_srpm_commits(upstream, downstream):
    package_config = upstream.package_config
    ChangelogHelper(upstream, downstream, package_config).prepare_upstream_locally(
        "0.1.0",
        "abc123a",
        True,
        None,
    )
    with upstream.specfile.sections() as sections:
        assert "- Development snapshot (abc123a)" in sections.changelog


def test_srpm_no_tags(upstream, downstream):
    package_config = upstream.package_config
    flexmock(upstream).should_receive("get_last_tag").and_return(None).once()

    ChangelogHelper(upstream, downstream, package_config).prepare_upstream_locally(
        "0.1.0",
        "abc123a",
        True,
        None,
    )
    with upstream.specfile.sections() as sections:
        assert "- Development snapshot (abc123a)" in sections.changelog


def test_srpm_no_bump(upstream, downstream):
    package_config = upstream.package_config
    flexmock(upstream).should_receive("get_last_tag").and_return(None).once()

    ChangelogHelper(upstream, downstream, package_config).prepare_upstream_locally(
        "0.1.0",
        "abc123a",
        False,
        None,
    )
    with upstream.specfile.sections() as sections:
        assert "- Development snapshot (abc123a)" not in sections.changelog


def test_update_distgit_when_copy_upstream_release_description(
    upstream,
    distgit_instance,
):
    _, downstream = distgit_instance
    package_config = upstream.package_config
    package_config.copy_upstream_release_description = True
    upstream.local_project.git_project = (
        flexmock()
        .should_receive("get_release")
        .with_args(tag_name="0.1.0", name="0.1.0")
        .and_return(flexmock(body="Some release 0.1.0"))
        .mock()
    )

    ChangelogHelper(upstream, downstream, package_config).update_dist_git(
        upstream_tag="0.1.0",
        full_version="0.1.0",
        resolved_bugs=["rhbz#123"],
    )

    with downstream._specfile.sections() as sections:
        assert "Some release 0.1.0" in sections.changelog
        assert "- Resolves: rhbz#123" in sections.changelog


@pytest.mark.skipif(
    rpm.__version__ < "4.16",
    reason="%autochangelog requires rpm 4.16 or higher",
)
def test_do_not_update_distgit_with_autochangelog(
    upstream,
    distgit_instance_with_autochangelog,
):
    _, downstream = distgit_instance_with_autochangelog
    package_config = upstream.package_config

    ChangelogHelper(upstream, downstream, package_config).update_dist_git(
        upstream_tag="0.1.0",
        full_version="0.1.0",
    )

    with downstream._specfile.sections() as sections:
        assert "%autochangelog" in sections.changelog


def test_update_distgit_unsafe_commit_messages(upstream, distgit_instance):
    _, downstream = distgit_instance
    package_config = upstream.package_config
    flexmock(upstream).should_receive("get_commit_messages").and_return(
        "* 100% of tests now pass\n"
        "* got rid of all shell (%(...)) and expression (%[...]) expansions\n"
        "* removed all %global macros\n"
        "* cleaned up %install section\n",
    )

    ChangelogHelper(upstream, downstream, package_config).update_dist_git(
        upstream_tag="0.1.0",
        full_version="0.1.0",
    )

    # make sure only one changelog entry (as seen by RPM) has been added
    downstream.specfile.macros.append(("_changelog_trimage", "0"))
    downstream.specfile.macros.append(("_changelog_trimtime", "0"))
    assert len(downstream.specfile.rpm_spec.sourceHeader[rpm.RPMTAG_CHANGELOGTEXT]) == 2


def test_update_distgit_when_copy_upstream_release_description_none(
    upstream,
    distgit_instance,
):
    _, downstream = distgit_instance
    package_config = upstream.package_config
    package_config.copy_upstream_release_description = True
    upstream.local_project.git_project = (
        flexmock()
        .should_receive("get_release")
        .with_args(tag_name="0.1.0", name="0.1.0")
        .and_return(flexmock(body=None))
        .mock()
    )

    ChangelogHelper(upstream, downstream, package_config).update_dist_git(
        upstream_tag="0.1.0",
        full_version="0.1.0",
    )

    with downstream._specfile.sections() as sections:
        assert "- Update to version 0.1.0" in sections.changelog


def test_update_distgit_changelog_entry_action_pass_env_vars(
    upstream,
    distgit_instance,
):
    _, downstream = distgit_instance
    package_config = upstream.package_config
    package_config.actions = {ActionName.changelog_entry: "command"}
    upstream.local_project.git_project = (
        flexmock()
        .should_receive("get_release")
        .with_args(tag_name="0.1.0", name="0.1.0")
        .and_return(flexmock(body="Some release 0.1.0"))
        .mock()
    )
    expected_env = {
        "PACKIT_CONFIG_PACKAGE_NAME": "beer",
        "PACKIT_UPSTREAM_PACKAGE_NAME": "beerware",
        "PACKIT_DOWNSTREAM_PACKAGE_NAME": "beer",
        "PACKIT_PROJECT_VERSION": "0.1.0",
        "PACKIT_RESOLVED_BUGS": "rhbz#123 rhbz#124",
        "PACKIT_PROJECT_UPSTREAM_TAG": "0.1.0",
        "PACKIT_PROJECT_PREVIOUS_VERSION": "0.0.0",
    }
    flexmock(upstream.actions_handler).should_receive(
        "get_output_from_action",
    ).with_args(
        ActionName.changelog_entry,
        env=expected_env,
    ).and_return(
        "- entry",
    ).once()

    ChangelogHelper(upstream, downstream, package_config).update_dist_git(
        upstream_tag="0.1.0",
        full_version="0.1.0",
        resolved_bugs=["rhbz#123", "rhbz#124"],
    )


def test_update_distgit_no_distgit_specfile(
    upstream,
    distgit_instance,
):
    d, downstream = distgit_instance
    # remove the downstream specfile
    d.joinpath("beer.spec").unlink()
    subprocess.check_call(
        ["git", "commit", "-m", "remove spec", "-a"],
        cwd=str(d),
    )
    package_config = upstream.package_config
    package_config.copy_upstream_release_description = True
    upstream.local_project.git_project = (
        flexmock()
        .should_receive("get_release")
        .with_args(tag_name="0.1.0", name="0.1.0")
        .and_return(flexmock(body="Some release 0.1.0"))
        .mock()
    )

    ChangelogHelper(upstream, downstream, package_config).update_dist_git(
        upstream_tag="0.1.0",
        full_version="0.1.0",
    )

    with downstream._specfile.sections() as sections:
        assert "Some release 0.1.0" in sections.changelog



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from packit.utils.changelog_helper import ChangelogHelper


@pytest.mark.parametrize(
    "original, sanitized",
    [
        ("- 100% of tests now pass", "- 100% of tests now pass"),
        ("- removed all %global macros", "- removed all %%global macros"),
        ("- cleaned up %install section", "- cleaned up %%install section"),
        (
            "- got rid of all shell (%(...)) and expression (%[...]) expansions",
            "- got rid of all shell (%%(...)) and expression (%%[...]) expansions",
        ),
        ("- first item\n* second item", "- first item\n * second item"),
        ("* first item\n* second item", " * first item\n * second item"),
    ],
)
def test_sanitize_entry(original, sanitized):
    assert ChangelogHelper.sanitize_entry(original) == sanitized



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from packit.utils.commands import run_command


def test_run_command_w_env():
    run_command(["bash", "-c", "env | grep PATH"], env={"X": "Y"})



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Generate initial configuration for packit
"""

import logging
import os
from pathlib import Path
from typing import Optional

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_existing_config
from packit.config import get_context_settings
from packit.config.config import pass_config
from packit.config.package_config import get_local_specfile_path
from packit.constants import PACKIT_CONFIG_TEMPLATE
from packit.exceptions import PackitException

logger = logging.getLogger(__name__)


@click.command("init", context_settings=get_context_settings())
@click.argument("path_or_url", type=LocalProjectParameter(), default=os.path.curdir)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Reset config to default if already exists.",
)
@pass_config
@cover_packit_exception
def init(
    config,
    path_or_url,
    force,
):
    """
    Create the initial Packit configuration in a repository

    See 'packit source-git init', if you want to initialize a repository
    as a source-git repo.
    """
    working_dir = path_or_url.working_dir
    config_path = get_existing_config(working_dir)
    if config_path:
        if not force:
            raise PackitException(
                f"Packit config {config_path} already exists."
                " If you want to regenerate it use `packit init --force`",
            )
    else:
        # Use default name
        config_path = working_dir / ".packit.yaml"

    specfile_path = get_local_specfile_path(working_dir)
    template_data = {
        "upstream_package_name": path_or_url.repo_name,
        "downstream_package_name": path_or_url.repo_name,
        "specfile_path": (
            specfile_path
            if specfile_path is not None
            else f"{path_or_url.repo_name}.spec"
        ),
    }

    with open(working_dir / ".gitignore", mode="a") as gitignore:
        print("prepare_sources_result*/", file=gitignore)

    generate_config(
        config_file=config_path,
        write_to_file=True,
        template_data=template_data,
    )


def generate_config(
    config_file: Path,
    write_to_file: bool = False,
    template_data: Optional[dict] = None,
) -> str:
    """
    Generate config file from provided data
    :param config_file: Path, .packit.yaml by default
    :param write_to_file: bool, write to config_file? False by default
    :param template_data: dict, example:
    {
        "upstream_package_name": "packitos",
        "downstream_package_name": "packit",
        "specfile_path": packit.spec,
    }
    :return: str, generated config
    """
    output_config = PACKIT_CONFIG_TEMPLATE.format(
        downstream_package_name=template_data["downstream_package_name"],
        upstream_package_name=template_data["upstream_package_name"],
        specfile_path=template_data["specfile_path"],
    )

    if write_to_file:
        config_file.write_text(output_config)
        logger.debug(f"Packit config file '{config_file}' changed.")

    return output_config



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Tests for Distgit class
"""

import pytest

from packit.exceptions import PackitException


def test_distgit_commit_empty(distgit_instance):
    d, dg = distgit_instance
    with pytest.raises(PackitException) as ex:
        dg.commit("", "")
    assert (
        str(ex.value)
        == "No changes are present in the dist-git repo: nothing to commit."
    )


def test_get_nvr(distgit_instance):
    d, dg = distgit_instance
    nvr = dg.get_nvr("main")
    assert nvr.startswith("dist_git_remote-0.0.0-1")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Validate PackageConfig
"""

import logging
import os

import click

from packit.api import PackitAPI
from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception
from packit.config import get_context_settings
from packit.local_project import LocalProject

logger = logging.getLogger(__name__)


@click.command("validate-config", context_settings=get_context_settings())
@click.argument("path_or_url", type=LocalProjectParameter(), default=os.path.curdir)
@click.option(
    "--offline",
    default=False,
    is_flag=True,
    help="Do not make remote API calls requiring network access.",
)
@cover_packit_exception
def validate_config(path_or_url: LocalProject, offline: bool):
    """
    Validate PackageConfig.

    \b
    - checks missing values
    - checks incorrect types
    - checks whether monitoring is enabled if 'pull_from_upstream` is used

    PATH_OR_URL argument is a local path or a URL to a git repository with packit configuration file
    """
    output = PackitAPI.validate_package_config(path_or_url.working_dir, offline)
    logger.info(output)
    # TODO: print more if config.debug



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import os

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api, iterate_packages
from packit.config import get_context_settings, pass_config
from packit.config.aliases import get_branches
from packit.constants import (
    DEFAULT_BODHI_NOTE,
    PACKAGE_LONG_OPTION,
    PACKAGE_OPTION_HELP,
    PACKAGE_SHORT_OPTION,
)
from packit.exceptions import PackitException

logger = logging.getLogger(__name__)


@click.command("create-update", context_settings=get_context_settings())
@click.option(
    "--dist-git-branch",
    help="Comma separated list of target branches in dist-git to create bodhi update in. "
    "(defaults to repo's default branch)",
)
@click.option(
    "--dist-git-path",
    help="Path to dist-git repo to work in. "
    "Otherwise clone the repo in a temporary directory.",
)
@click.option(
    "--koji-build",
    help="Koji build (NVR) to add to the bodhi update (can be specified multiple times)",
    required=False,
    multiple=True,
)
# It would make sense to open an editor here,
# just like `git commit` and get notes like that
@click.option(
    "--update-notes",
    help="Bodhi update notes",
    required=False,
    default=DEFAULT_BODHI_NOTE,
)
@click.option(
    "--update-type",
    type=click.types.Choice(("security", "bugfix", "enhancement", "newpackage")),
    help="Type of the bodhi update",
    required=False,
    default="enhancement",
)
@click.option(
    "-b",
    "--resolve-bug",
    help="Bugzilla IDs that are resolved with the update",
    required=False,
    multiple=True,
    type=click.INT,
)
@click.option(
    PACKAGE_SHORT_OPTION,
    PACKAGE_LONG_OPTION,
    multiple=True,
    help=PACKAGE_OPTION_HELP.format(action="update"),
)
@click.argument("path_or_url", type=LocalProjectParameter(), default=os.path.curdir)
@pass_config
@cover_packit_exception
@iterate_packages
def create_update(
    config,
    dist_git_branch,
    dist_git_path,
    koji_build,
    update_notes,
    update_type,
    resolve_bug,
    package_config,
    path_or_url,
):
    """
    Create a bodhi update for the selected upstream project

    If you are not authenticated with the bodhi server, please make sure that you
    navigate in your browser to the URL provided by the bodhi-client and then paste
    the `code=XX...` to the terminal when prompted.

    If you set `fas_user` and `kerberos_realm` in your "~/.config/packit.yaml" and
    have an active Kerberos TGT, you will be automatically authenticated. Otherwise,
    you need to follow the prompt

    PATH_OR_URL argument is a local path or a URL to the upstream git repository,
    it defaults to the current working directory
    """
    api = get_packit_api(
        config=config,
        package_config=package_config,
        dist_git_path=dist_git_path,
        local_project=path_or_url,
    )
    default_dg_branch = api.dg.local_project.git_project.default_branch
    dist_git_branch = dist_git_branch or default_dg_branch
    branches_to_update = get_branches(
        *dist_git_branch.split(","),
        default_dg_branch=default_dg_branch,
    )
    click.echo(
        f"Creating Bodhi update for the following branches: {', '.join(branches_to_update)}",
    )

    if branches_to_update:
        click.echo("Please provide Bodhi username and password when asked for.")

    for branch in branches_to_update:
        try:
            api.create_update(
                koji_builds=koji_build,
                dist_git_branch=branch,
                update_notes=update_notes,
                update_type=update_type,
                bugzilla_ids=resolve_bug,
            )
        except PackitException as ex:  # noqa: PERF203
            click.echo(
                f"There was a problem while creating an update for {branch}:\n{ex}\n\n"
                "Please try again later if this looks like a transient issue.\n"
                "If you believe this is a bug, please contact Fedora infrastructure:\n"
                "  https://pagure.io/fedora-infrastructure\n\n"
                "You are always welcome to contact the Packit team:\n"
                "  https://github.com/packit/packit/issues",
                err=True,
            )
            return



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
import logging
import os
import re
import shlex
import shutil
import tarfile
import tempfile
from functools import partial, reduce
from pathlib import Path
from typing import Optional, Union

import git
import rpm
from lazy_object_proxy import Proxy
from ogr.services.pagure import PagureProject
from specfile import Specfile

from packit.actions import ActionName
from packit.actions_handler import ActionsHandler
from packit.base_git import PackitRepositoryBase
from packit.command_handler import (
    RUN_COMMAND_HANDLER_MAPPING,
    CommandHandler,
    SandcastleCommandHandler,
)
from packit.config import Config
from packit.config.common_package_config import MultiplePackages
from packit.constants import DATETIME_FORMAT, DEFAULT_ARCHIVE_EXT
from packit.distgit import DistGit
from packit.exceptions import (
    PackitCommandFailedError,
    PackitException,
    PackitFailedToCreateRPMException,
    PackitFailedToCreateSRPMException,
    PackitRPMNotFoundException,
    PackitSRPMNotFoundException,
)
from packit.local_project import LocalProject
from packit.patches import PatchGenerator, PatchMetadata
from packit.sync import iter_srcs
from packit.utils import commands, sanitize_version
from packit.utils.changelog_helper import ChangelogHelper
from packit.utils.commands import run_command
from packit.utils.repo import get_current_version_command, git_remote_url_to_https_url
from packit.utils.upstream_version import get_upstream_version
from packit.utils.versions import compare_versions

logger = logging.getLogger(__name__)


class Upstream:
    """Interact with upstream project"""

    def __init__(
        self,
        config: Config,
        package_config: MultiplePackages,
    ):
        """
        Args:
            config: global configuration
            package_config: configuration of the upstream project
        """
        self.config = config
        self.package_config = package_config
        self.allowed_gpg_keys: Optional[list[str]] = None

        self._handler_kls = None
        self._command_handler: Optional[CommandHandler] = None
        self._actions_handler: Optional[ActionsHandler] = None
        self._specfile: Optional[Specfile] = None

    def __repr__(self):
        return (
            "Upstream("
            f"config='{self.config}', "
            f"package_config='{self.package_config}')"
        )

    @property
    def absolute_specfile_dir(self) -> Optional[Path]:
        raise NotImplementedError()

    @property
    def specfile(self) -> Optional[Specfile]:
        return self._specfile

    @property
    def local_project(self) -> Optional[LocalProject]:
        raise NotImplementedError()

    @property
    def working_dir(self) -> Optional[Path]:
        raise NotImplementedError()

    @property
    def handler_kls(self):
        if self._handler_kls is None:
            logger.debug(f"Command handler: {self.config.command_handler}")
            self._handler_kls = RUN_COMMAND_HANDLER_MAPPING[self.config.command_handler]
        return self._handler_kls

    @property
    def command_handler(self) -> CommandHandler:
        raise NotImplementedError()

    @property
    def actions_handler(self) -> ActionsHandler:
        if not self._actions_handler:
            self._actions_handler = ActionsHandler(
                self.package_config,
                self.command_handler,
            )
        return self._actions_handler

    @property
    def commit_hexsha(self) -> Optional[str]:
        raise NotImplementedError()

    @property
    def active_branch(self) -> Optional[str]:
        raise NotImplementedError()

    @property
    def absolute_specfile_path(self) -> Optional[Path]:
        raise NotImplementedError()

    @staticmethod
    def _template2regex(template) -> str:
        """
        Converts tag template to regex with named groups.

        Args:
            template: tag_template string
        Returns:
            regex string which can be used by python re module
        """

        p = re.compile(r"{(.*?)}")
        return p.sub(r"(?P<\g<1>>.*)", template)

    def get_latest_released_version(self) -> Optional[str]:
        """
        Return version of the upstream project for the latest official release

        Returns:
            the version string (e.g. "1.0.0")
        """

        version = get_upstream_version(self.package_config.downstream_package_name)
        logger.info(
            f"Version retrieved from release-monitoring.org is {version!r}.",
        )
        return version

    def get_version_from_action(self) -> Optional[str]:
        """Provide version from action"""
        action_output = self.actions_handler.get_output_from_action(
            action=ActionName.get_current_version,
            env=self.package_config.get_package_names_as_env(),
        )
        return action_output[-1].strip() if action_output else None

    def get_version_from_tag(self, tag: Optional[str]) -> Optional[str]:
        """
        Extracts version from git tag using `upstream_template_tag`.

        Args:
            tag: Git tag containing version.

        Returns:
            Version string or `None` if given empty string or `None`.
        """
        if not tag:
            return None

        field = "version"
        regex = self._template2regex(self.package_config.upstream_tag_template)
        p = re.compile(regex)
        match = p.match(tag)

        if match and field in match.groupdict():
            return match.group(field)

        msg = (
            f'Unable to extract "{field}" from {tag} using '
            f"{self.package_config.upstream_tag_template}"
        )
        logger.error(msg)
        raise PackitException(msg)

    def convert_version_to_tag(self, version_: str) -> str:
        """
        Converts version to tag using upstream_tag_tepmlate

        Args:
            version_: version to be converted
            upstream_template_tag

        Returns:
             tag
        """
        try:
            tag = self.package_config.upstream_tag_template.format(version=version_)
        except KeyError as e:
            msg = (
                f"Invalid upstream_tag_template: {self.package_config.upstream_tag_template} - "
                f'"version" placeholder is missing'
            )
            logger.error(msg)
            raise PackitException(msg) from e

        return tag

    def set_specfile(self, specfile: Specfile):
        self._specfile = specfile

    def get_specfile_version(self) -> str:
        """provide version from specfile"""
        # we need to get the version from rpm spec header
        # (as the version tag might not be present directly in the specfile,
        # but e.g. imported)
        version = self.specfile.rpm_spec.sourceHeader[rpm.RPMTAG_VERSION]
        logger.info(f"Version in spec file is {version!r}.")
        return version

    def is_command_handler_set(self) -> bool:
        """return True when command_handler is initialized"""
        return bool(self._command_handler)

    def get_absolute_specfile_path(self) -> Optional[Path]:
        raise NotImplementedError()

    def sync_files(self, synced_files: list, dg: DistGit):
        # Make all paths absolute and check that they are within
        # the working directories of the repositories.
        for item in synced_files:
            item.resolve(
                src_base=self.working_dir,
                dest_base=dg.local_project.working_dir,
            )

    def _expand_git_ref(self, ref: Optional[str]) -> Optional[str]:
        raise NotImplementedError()

    def is_dirty(self):
        raise NotImplementedError()

    def create_patches(
        self,
        upstream: Optional[str] = None,
        destination: Optional[Union[str, Path]] = None,
    ) -> list[PatchMetadata]:
        raise NotImplementedError()

    def get_last_tag(self, before: Optional[str] = None) -> Optional[str]:
        raise NotImplementedError()

    def checkout_release(self, upstream_tag: str):
        raise NotImplementedError()

    def check_last_commit(self):
        raise NotImplementedError()

    def create_rpms(self, rpm_dir: Union[str, Path, None] = None) -> list[Path]:
        raise NotImplementedError()

    def push_to_fork(
        self,
        branch_name: str,
        force: bool = False,
        fork: bool = True,
        remote_name: Optional[str] = None,
        sync_acls: Optional[bool] = False,
    ) -> tuple[str, Optional[str]]:
        raise NotImplementedError()

    def create_pull(
        self,
        pr_title: str,
        pr_description: str,
        source_branch: str,
        target_branch: str,
        fork_username: Optional[str] = None,
    ) -> None:
        raise NotImplementedError()

    def create_branch(
        self,
        branch_name: str,
        base: str = "HEAD",
        setup_tracking: bool = False,
    ) -> git.Head:
        raise NotImplementedError()

    def switch_branch(
        self,
        branch: Optional[str] = None,
        force: Optional[bool] = False,
    ) -> None:
        raise NotImplementedError()

    def prepare_upstream_for_srpm_creation(
        self,
        upstream_ref: Optional[str] = None,
        update_release: bool = True,
        release_suffix: Optional[str] = None,
        create_symlinks: Optional[bool] = True,
        merged_ref: Optional[str] = None,
    ):
        raise NotImplementedError()

    def push(self, refspec: str, remote_name: str = "origin", force: bool = False):
        raise NotImplementedError()

    def get_commit_messages(
        self,
        after: Optional[str] = None,
        before: str = "HEAD",
    ) -> str:
        raise NotImplementedError()

    def koji_build(
        self,
        scratch: bool = False,
        nowait: bool = False,
        koji_target: Optional[str] = None,
        srpm_path: Optional[Path] = None,
    ):
        raise NotImplementedError()

    def commit(
        self,
        title: str,
        msg: str,
        prefix: str = "[packit] ",
        trailers: Optional[list[tuple[str, str]]] = None,
    ) -> None:
        raise NotImplementedError()

    def create_srpm(
        self,
        srpm_path: Union[Path, str, None] = None,
        srpm_dir: Union[Path, str, None] = None,
    ) -> Path:
        raise NotImplementedError()

    def clean_working_dir(self):
        raise NotImplementedError()


class NonGitUpstream(Upstream):
    """Interact with non-git upstream project"""

    def __init__(
        self,
        config: Config,
        package_config: MultiplePackages,
    ):
        super().__init__(config=config, package_config=package_config)
        self._working_dir: Optional[Path] = None

    def __repr__(self):
        return (
            "NonGitUpstream("
            f"config='{self.config}', "
            f"package_config='{self.package_config}')"
        )

    @property
    def absolute_specfile_dir(self) -> Optional[Path]:
        return None

    @property
    def local_project(self) -> Optional[LocalProject]:
        return None

    @property
    def working_dir(self) -> Optional[Path]:
        if not self._working_dir:
            if self.handler_kls == SandcastleCommandHandler:
                path = (
                    Path(self.config.command_handler_work_dir) / "non-git-working-dir"
                )
                path.mkdir(parents=True, exist_ok=True)
                self._working_dir = path
            else:
                self._working_dir = Path(tempfile.mkdtemp())
            logger.info(
                f"Created directory for actions and syncing: {self._working_dir}",
            )
        return self._working_dir

    @property
    def command_handler(self) -> CommandHandler:
        if self._command_handler is None:
            self._command_handler = self.handler_kls(
                config=self.config,
                working_dir=self.working_dir,
            )
        return self._command_handler

    @property
    def commit_hexsha(self) -> Optional[str]:
        return None

    @property
    def active_branch(self) -> Optional[str]:
        return None

    @property
    def absolute_specfile_path(self) -> Optional[Path]:
        return None

    def get_absolute_specfile_path(self) -> Optional[Path]:
        return None

    def _expand_git_ref(self, ref: Optional[str]) -> Optional[str]:
        return None

    def is_dirty(self):
        return False

    def create_patches(
        self,
        upstream: Optional[str] = None,
        destination: Optional[Union[str, Path]] = None,
    ) -> list[PatchMetadata]:
        pass

    def get_last_tag(self, before: Optional[str] = None) -> Optional[str]:
        return None

    def checkout_release(self, upstream_tag: str):
        pass

    def clean_working_dir(self):
        if not self._working_dir:
            return

        logger.debug(f"Cleaning: {self.working_dir}")
        shutil.rmtree(self.working_dir, ignore_errors=True)


class GitUpstream(PackitRepositoryBase, Upstream):
    """Interact with git upstream project"""

    def __init__(
        self,
        config: Config,
        package_config: MultiplePackages,
        local_project: LocalProject,
    ):
        """
        Args:
            config: global configuration
            package_config: configuration of the upstream project
            local_project: public offender
        """
        self._local_project = local_project
        super().__init__(config=config, package_config=package_config)
        self.config = config
        self.package_config = package_config
        self._project_required = True
        self._merged_ref: Optional[str] = None

    def __repr__(self):
        return (
            "GitUpstream("
            f"config='{self.config}', "
            f"package_config='{self.package_config}', "
            f"local_project='{self.local_project}', "
            f"active_branch='{self.active_branch}')"
        )

    @property
    def local_project(self):
        if not self._local_project:
            self._local_project = LocalProject(
                git_url=self.package_config.upstream_project_url,
                repo_name=self.package_config.upstream_package_name,
                cache=self.repository_cache,
                merge_pr=self.package_config.merge_pr_in_ci,
            )
            # TODO: Turn this on once p-s mocks are updated
            # builder = LocalProjectBuilder(cache=self.repository_cache)
            # self._local_project = builder.build(
            #    git_url=self.package_config.upstream_project_url,
            #    repo_name=self.package_config.upstream_package_name,
            #    merge_pr=self.package_config.merge_pr_in_ci,
            #    git_repo=CALCULATE,
            #    working_dir=CALCULATE,
            #    ref=CALCULATE,
            #    git_project=CALCULATE,
            # )
        if self._local_project.git_project is None:
            if not self.package_config.upstream_project_url:
                self.package_config.upstream_project_url = git_remote_url_to_https_url(
                    self._local_project.git_url,
                )

            self._local_project.git_project = self.config.get_project(
                url=self.package_config.upstream_project_url,
                required=self._project_required,
            )
        return self._local_project

    @property
    def working_dir(self) -> Optional[Path]:
        return self.local_project.working_dir

    @property
    def commit_hexsha(self) -> str:
        return self.local_project.commit_hexsha

    @property
    def command_handler(self) -> CommandHandler:
        if self._command_handler is None:
            self._command_handler = self.handler_kls(
                # so that the local_project is evaluated only when needed
                local_project=Proxy(partial(GitUpstream.local_project.__get__, self)),  # type: ignore
                config=self.config,
            )
        return self._command_handler

    @property
    def active_branch(self) -> str:
        return self.local_project.ref

    def clean_working_dir(self):
        pass

    def checkout_release(self, upstream_tag: str):
        self.local_project.checkout_release(upstream_tag)

    def push_to_fork(
        self,
        branch_name: str,
        force: bool = False,
        fork: bool = True,
        remote_name: Optional[str] = None,
        sync_acls: Optional[bool] = False,
    ) -> tuple[str, Optional[str]]:
        """
        Push current branch to fork if fork=True, else to origin.

        Args:
            branch_name: the branch where we push
            force: push forcefully?
            fork: push to fork?
            remote_name: name of remote where we should push
               if None, try to find a ssh_url
            sync_acls: whether to sync the ACLs of the original repo and the fork

        Returns:
            name of the branch where we pushed

        """
        logger.debug(
            f"About to {'force ' if force else ''}push changes to branch {branch_name!r}.",
        )
        fork_username = None

        if not remote_name:
            if fork:
                if self.local_project.git_project.is_fork:
                    project = self.local_project.git_project
                else:
                    # ogr is awesome! if you want to fork your own repo, you'll get it!
                    project = self.local_project.git_project.get_fork(create=True)

                if sync_acls and isinstance(
                    self.local_project.git_project,
                    PagureProject,
                ):
                    # synchronize ACLs between original repo and fork for Pagure
                    self.sync_acls(self.local_project.git_project, project)

                fork_username = project.namespace
                fork_urls = project.get_git_urls()

                ssh_url = fork_urls["ssh"]

                remote_name = "fork-ssh"
                for remote in self.local_project.git_repo.remotes:
                    pushurl = next(remote.urls)  # afaik this is what git does as well
                    if ssh_url.startswith(pushurl):
                        logger.info(
                            f"Will use remote {remote!r} using URL {pushurl!r}.",
                        )
                        remote_name = str(remote)
                        break
                else:
                    logger.info(f"Creating remote fork-ssh with URL {ssh_url!r}.")
                    self.local_project.git_repo.create_remote(
                        name="fork-ssh",
                        url=ssh_url,
                    )
            else:
                # push to origin and hope for the best
                remote_name = "origin"
        logger.info(f"Pushing to remote {remote_name!r} using branch {branch_name!r}.")
        try:
            self.push(refspec=branch_name, force=force, remote_name=remote_name)
        except git.GitError as ex:
            msg = (
                f"Unable to push to remote {remote_name!r} using branch {branch_name!r}, "
                f"the error is:\n{ex}"
            )
            raise PackitException(msg) from ex
        return str(branch_name), fork_username

    def create_pull(
        self,
        pr_title: str,
        pr_description: str,
        source_branch: str,
        target_branch: str,
        fork_username: Optional[str] = None,
    ) -> None:
        """
        Create upstream pull request using the requested branches
        """
        project = self.local_project.git_project

        try:
            upstream_pr = project.create_pr(
                title=pr_title,
                body=pr_description,
                source_branch=source_branch,
                target_branch=target_branch,
                fork_username=fork_username,
            )
        except Exception as ex:
            logger.error(f"There was an error while creating a PR: {ex!r}")
            raise
        else:
            logger.info(f"PR created: {upstream_pr.url}")

    def create_patches(
        self,
        upstream: Optional[str] = None,
        destination: Optional[Union[str, Path]] = None,
    ) -> list[PatchMetadata]:
        """
        Create patches from downstream commits.

        Args:
            destination: str
            upstream: str -- git branch or tag

        Returns:
             [PatchMetadata, ...] list of patches
        """
        upstream = upstream or self.get_specfile_version()
        destination = Path(destination) or self.local_project.working_dir

        sync_files_to_ignore = self.package_config.get_all_files_to_sync()
        for file in sync_files_to_ignore:
            file.resolve(
                src_base=self.local_project.working_dir,
                # dest (downstream) is not important, we only care about src (upstream)
                dest_base=destination,
            )
        sync_files_to_ignore = [
            str(Path(file).relative_to(self.local_project.working_dir))
            for file in iter_srcs(sync_files_to_ignore)
        ]
        files_to_ignore = (
            self.package_config.patch_generation_ignore_paths + sync_files_to_ignore
        )

        pg = PatchGenerator(self.local_project)
        return pg.create_patches(
            upstream,
            str(destination),
            files_to_ignore=files_to_ignore,
        )

    def get_version(self) -> str:
        """
        Return version given by action, if any.
        Or the latest release available: prioritize bigger from upstream
        package repositories or the version in spec
        """
        if action_version := self.get_version_from_action():
            return action_version

        ups_ver = self.get_latest_released_version()
        spec_ver = self.get_specfile_version()

        if ups_ver and compare_versions(ups_ver, spec_ver) > 0:
            logger.warning(f"Version {spec_ver!r} in spec file is outdated.")
            logger.info(f"Picking version {ups_ver!r} from release-monitoring.org.")
            return ups_ver

        logger.info(f"Picking version {spec_ver!r} found in spec file.")
        return spec_ver

    def get_current_version(self) -> str:
        """
        Get version of the project in current state. Tries following steps:

        1. Get output from actions
        2. Extract version from `self.get_last_tag()` using `self.get_version_from_tag()`
        3. Falls back to version in the specfile.

        Returns:
            String containing version, e.g. `"0.1.1"`.
        """
        # Step 1
        version = self.get_version_from_action()

        # Step 2
        if version is None:
            version = self.get_version_from_tag(self.get_last_tag())

        # Step 3
        if version is None:
            logger.info("No git tags found, falling back to %version in the specfile.")
            version = self.specfile.expanded_version

        logger.debug(f"Version: {version}")
        version = sanitize_version(version)
        logger.debug(f"Sanitized version: {version}")

        return version

    def create_archive(
        self,
        version: Optional[str] = None,
        create_symlink: Optional[bool] = True,
    ) -> str:
        return Archive(self, version).create(create_symlink=create_symlink)

    def list_tags(self, merged_ref: Optional[str] = None) -> list[str]:
        """
        List tags in the repository sorted by created date from the most recent.

        Args:
            merged_ref: List only tags reachable from this git ref.

        Returns:
            List of tags.
        """
        try:
            cmd = [
                "git",
                "tag",
                "--list",
                "--sort=-creatordate",
            ]
            if merged_ref is not None:
                cmd.append(f"--merged={merged_ref}")
            tags = run_command(
                cmd,
                output=True,
                cwd=self.local_project.working_dir,
            ).stdout.split()
        except PackitCommandFailedError as ex:
            logger.debug(f"{ex!r}")
            logger.info("Can't list the tags in this repository.")
            return []

        return tags

    def get_last_tag(self, before: Optional[str] = None) -> Optional[str]:
        """
        Get last git-tag (matching the configuration) from the repo.

        Args:
            before: get the last tag before this tag

        Returns:
            Last matching tag.
        """

        logger.debug(
            f"We're about to get latest matching tag in "
            f"the upstream repository {self.local_project.working_dir}.",
        )

        tags = self.list_tags(self._merged_ref)

        if not tags:
            logger.info("No tags found in the repository.")
            return None

        if before:
            if before not in tags:
                logger.debug(f"{before} not present in the obtained list of tags.")
                return None
            index = tags.index(before)
            tags = tags[(index + 1) :]

        matching_tags = self.filter_tags(tags)
        return matching_tags[0] if matching_tags else None

    def filter_tags(self, tags: list[str]):
        """
        Filter the given tags using `upstream_tag_include` and
        `upstream_tag_exclude` if they are present.

        Args:
            tags: list of tags that should be filtered

        Returns:
            Tags that match with `upstream_tag_include` and
            `upstream_tag_exclude`.
        """
        if self.package_config.upstream_tag_include:
            include_pattern = re.compile(self.package_config.upstream_tag_include)
            tags = [tag for tag in tags if include_pattern.match(tag)]
            logger.debug(f"Filtered tags after matching upstream_tag_include: {tags}")

        if self.package_config.upstream_tag_exclude:
            exclude_pattern = re.compile(self.package_config.upstream_tag_exclude)
            tags = [tag for tag in tags if not exclude_pattern.match(tag)]
            logger.debug(f"Filtered tags after matching upstream_tag_exclude: {tags}")

        return tags

    def get_commit_messages(
        self,
        after: Optional[str] = None,
        before: str = "HEAD",
    ) -> str:
        """
        Args:
            after: get commit messages after this revision,
                    if None, all commit messages before 'before' will be returned
            before:  get commit messages before this revision

        Returns:
            commit messages
        """
        # let's print changes b/w the last 2 revisions;
        # ambiguous argument '0.1.0..HEAD': unknown revision or path not in the working tree.
        # Use '--' to separate paths from revisions, like this
        commits_range = f"{after}..{before}" if after else before
        if not before:
            raise PackitException(
                "Unable to get a list of commit messages in range "
                f"{commits_range} because the upper bound is not "
                f"defined ({before!r}).",
            )
        cmd = [
            "git",
            "log",
            "--no-merges",
            "--pretty=format:- %s (%an)",
            commits_range,
            "--",
        ]
        try:
            return run_command(
                cmd,
                output=True,
                cwd=self.local_project.working_dir,
            ).stdout.strip()
        except PackitCommandFailedError as ex:
            logger.error(f"We couldn't get commit messages for %changelog\n{ex}")
            logger.info(f"Does the git ref {after} exist in the git repo?")
            logger.info(
                "If the ref is a git tag, "
                'you should consider setting "upstream_tag_template":\n  '
                "https://packit.dev/docs/configuration/#upstream_tag_template",
            )
            raise

    def get_spec_release(self, release_suffix: Optional[str] = None) -> Optional[str]:
        """Assemble pieces of the spec file %release field we intend to set
        within the default fix-spec-file action

        The format is:
            {original_release_number}.{current_time}.{sanitized_current_branch}{git_desc_suffix}

        Example:
            1.20210913173257793557.packit.experiment.24.g8b618e91

        Returns:
            string which is meant to be put into a spec file %release field by packit
        """
        original_release_number = self.specfile.expanded_release.split(".", 1)[0]

        if release_suffix:
            return f"{original_release_number}.{release_suffix}"

        # we only care about the first number in the release
        # so that we can re-run `packit srpm`
        git_des_command = [
            "git",
            "describe",
            "--tags",
            "--long",
            "--match",
            "*",
        ]
        try:
            git_des_out = run_command(
                git_des_command,
                output=True,
                cwd=self.local_project.working_dir,
            ).stdout.strip()
        except PackitCommandFailedError as ex:
            # probably no tags in the git repo
            logger.info(f"Exception while describing the repository: {ex!r}")
            git_desc_suffix = ""
        else:
            # git adds various info in the output separated by -
            # so let's just drop version and reuse everything else
            g_desc_raw = git_des_out.rsplit("-", 2)[1:]
            # release components are meant to be separated by ".", not "-"
            git_desc_suffix = "." + ".".join(g_desc_raw)
            # the leading dot is put here b/c git_desc_suffix can be empty
            # and we could have two subsequent dots - rpm errors out in such a case
        current_branch = self.local_project.ref
        sanitized_current_branch = sanitize_version(current_branch)
        current_time = datetime.datetime.now().strftime(DATETIME_FORMAT)
        return (
            f"{original_release_number}.{current_time}."
            f"{sanitized_current_branch}{git_desc_suffix}"
        )

    def fix_spec(
        self,
        archive: str,
        version: str,
        commit: str,
        release: str,
        update_release: bool = True,
    ):
        """
        In order to create a SRPM from current git checkout, we need to have the spec reference
        the tarball and unpack it. This method updates the spec so it's possible.

        Args:
            archive: Relative path to the archive, used as `Source0`.
            version: Version to be set in the spec-file.
            commit: Commit to be set in the changelog.
            release: Release to be set in the spec-file.
            update_release: Whether to change Release in the spec-file.
                Defaults to `True`.
        """
        self._fix_spec_source(archive)
        self._fix_spec_prep(archive)

        ChangelogHelper(
            self,
            package_config=self.package_config,
        ).prepare_upstream_locally(
            version,
            commit,
            update_release,
            release,
        )

    def _fix_spec_prep(self, archive):
        with self.specfile.prep() as prep:
            if not prep:
                logger.warning("This package doesn't have a %prep section.")
                return

            if "%setup" in prep:
                macro = prep.setup
            elif "%autosetup" in prep:
                macro = prep.autosetup
            else:
                logger.warning(
                    "This package is not using %(auto)setup macro in prep. "
                    "Packit will not update the %prep section.",
                )
                return

            archive_root_dir = Archive(self, self.get_version()).get_archive_root_dir(
                archive,
            )

            if "n" in macro.options:
                # replace -n with our -n because it's better
                macro.options.n = archive_root_dir

    def _fix_spec_source(self, archive):
        number = self.package_config.spec_source_id_number
        with self.specfile.sources() as sources:
            source = next((s for s in sources if s.number == number), None)
            if source:
                source.location = archive
            else:
                raise PackitException(
                    "The spec file doesn't have sources set "
                    f"via {self.package_config.spec_source_id} nor Source.",
                )

    def create_srpm(
        self,
        srpm_path: Union[Path, str, None] = None,
        srpm_dir: Union[Path, str, None] = None,
    ) -> Path:
        """
        Create SRPM from the actual content of the repo.

        Args:
            srpm_path (Union[Path, str]): Path to the SRPM.

                Defaults to `None`.
            srpm_dir (Union[Path, str]): Path to the directory where the SRPM is
                meant to be placed.

                Defaults to `None`.

        Returns:
            Path to the SRPM.
        """
        return SRPMBuilder(
            upstream=self,
            srpm_path=srpm_path,
            srpm_dir=srpm_dir,
        ).build()

    def prepare_upstream_for_srpm_creation(
        self,
        upstream_ref: Optional[str] = None,
        update_release: bool = True,
        release_suffix: Optional[str] = None,
        create_symlinks: Optional[bool] = True,
        merged_ref: Optional[str] = None,
    ):
        """
        1. determine version
        2. create an archive or download upstream and create patches for sourcegit
        3. fix/update the specfile to use the right archive
        4. download the remote sources

        Args:
            upstream_ref: the base git ref for the source git
            update_release: update Release in spec file
            release_suffix: suffix %release part of NVR with this
            create_symlinks: whether symlinks should be created instead of copying the files
                (e.g. when the archive is created outside the specfile dir)
            merged_ref: git ref in the upstream repo used to identify correct most recent tag
        """
        try:
            self._merged_ref = merged_ref
            SRPMBuilder(upstream=self, ref=upstream_ref).prepare(
                update_release=update_release,
                release_suffix=release_suffix,
                create_symlinks=create_symlinks,
            )
        finally:
            self._merged_ref = None

    def create_patches_and_update_specfile(self, upstream_ref) -> None:
        """
        Create patches for the sourcegit and add them to the specfile.

        Args:
            upstream_ref: the base git ref for the source git
        """
        env = self.package_config.get_package_names_as_env()
        if self.actions_handler.with_action(action=ActionName.create_patches, env=env):
            patches = self.create_patches(
                upstream=upstream_ref,
                destination=str(self.absolute_specfile_dir),
            )
            self.specfile_add_patches(patches)

    def koji_build(
        self,
        scratch: bool = False,
        nowait: bool = False,
        koji_target: Optional[str] = None,
        srpm_path: Optional[Path] = None,
    ):
        """
        Perform a `koji build` in the repository

        Args:
            scratch: should the build be a scratch build?
            nowait: don't wait on build?
            koji_target: koji target to pick (see `koji list-targets`)
            srpm_path: use selected SRPM for build, not dist-git repo & ref
        """
        if not koji_target:
            raise PackitException(
                "koji target needs to be set when building directly from upstream",
            )
        # we can't use fedpkg b/c upstream repo is not dist-git
        cmd = shlex.split(self.config.koji_build_command)
        if scratch:
            cmd.append("--scratch")
        if nowait:
            cmd.append("--nowait")
        cmd += [koji_target, str(srpm_path)]
        logger.info("Starting a koji build.")
        if not nowait:
            logger.info(
                "We will be actively waiting for the build to finish, it may take some time.",
            )
        return commands.run_command_remote(
            cmd,
            cwd=self.local_project.working_dir,
            output=True,
            print_live=True,
        ).stdout

    def _build_rpms(
        self,
        mode: str,
        rpm_dir: Union[str, Path],
        rpmbuild_dir: Union[str, Path],
        src_dir: Union[str, Path],
        source: Union[str, Path],
    ) -> list[Path]:
        """
        Wrapper for building RPMs either from SRPM or specfile.

        Args:
            mode:
            rpm_dir: Path to the directory where the RPMs are meant to be placed.
            rpmbuild_dir: Path to the directory used during an RPM build.
            src_dir: Path to the directory with sources.
            source: Path to the source used for the build. Either SRPM or specfile.

        Returns:
            Paths to the built RPMs.
        """
        rpm_dir = rpm_dir or os.getcwd()
        src_dir = rpmbuild_dir = str(self.absolute_specfile_dir)

        cmd = [
            "rpmbuild",
            mode,
            "--define",
            f"_sourcedir {rpmbuild_dir}",
            "--define",
            f"_srcdir {src_dir}",
            "--define",
            f"_specdir {rpmbuild_dir}",
            "--define",
            f"_topdir {rpmbuild_dir}",
            "--define",
            f"_builddir {rpmbuild_dir}",
            "--define",
            f"_rpmdir {rpm_dir}",
            "--define",
            f"_buildrootdir {rpmbuild_dir}",
            str(source),
        ]

        escaped_command = " ".join(cmd)
        logger.debug(f"RPM build command: {escaped_command}")
        try:
            out = self.command_handler.run_command(
                cmd,
                return_output=True,
            ).stdout.strip()
        except PackitCommandFailedError as ex:
            logger.error(f"The `rpmbuild` command failed: {ex!r}")
            raise PackitFailedToCreateRPMException(
                f"reason:\n"
                f"{ex}\n"
                f"command:\n"
                f"{escaped_command}\n"
                f"stdout:\n"
                f"{ex.stdout_output}\n"
                f"stderr:\n"
                f"{ex.stderr_output}",
            ) from ex
        except PackitException as ex:
            logger.error(f"The `rpmbuild` command failed: {ex!r}")
            raise PackitFailedToCreateRPMException(
                f"The `rpmbuild` command failed:\n{ex}",
            ) from ex

        rpms = GitUpstream._get_rpms_from_rpmbuild_output(out)
        return [Path(rpm) for rpm in rpms]

    def create_rpms(self, rpm_dir: Union[str, Path, None] = None) -> list[Path]:
        """
        Create RPMs from the actual content of the repo.

        Args:
            rpm_dir: Path to the directory where the RPMs are meant to be placed.

                Defaults to current working directory, if not given.

        Returns:
            List of paths to the built RPMs.
        """
        rpm_dir = rpm_dir or os.getcwd()
        src_dir = rpmbuild_dir = str(self.absolute_specfile_dir)
        return self._build_rpms(
            "-bb",
            rpm_dir,
            rpmbuild_dir,
            src_dir,
            self.package_config.specfile_path,
        )

    def create_rpms_from_srpm(
        self,
        srpm: Union[str, Path],
        rpm_dir: Union[str, Path, None] = None,
    ) -> list[Path]:
        """
        Build RPMs from the given path to the SRPM.

        Args:
            srpm: Path to the SRPM to be built.
            rpm_dir: Path to the directory where the RPMs are meant to be placed.

                Defaults to current working directory, if not given.

        Returns:
            Paths to the built RPMs.
        """
        rpm_dir = rpm_dir or os.getcwd()
        return self._build_rpms("-rb", rpm_dir, rpm_dir, rpm_dir, srpm)

    @staticmethod
    def _get_rpms_from_rpmbuild_output(output: str) -> list[str]:
        """
        Try to find the rpm files in the `rpmbuild -bb` command output.

        Args:
            output: Output of the `rpmbuild -bb` command.

        Returns:
            List of names of the RPM files.
        """
        logger.debug(f"The `rpmbuild` command output: {output}")
        reg = r": (\S+\.rpm)(\s|$)"
        rpms = re.findall(reg, output)

        rpms = [rpm for rpm, _ in rpms]
        logger.debug(rpms)

        if not rpms:
            raise PackitRPMNotFoundException(
                "RPMs cannot be found, something is wrong.",
            )

        return rpms

    def _expand_git_ref(self, ref: Optional[str]) -> str:
        """
        Expands given git ref.
        If given globbing pattern, tries to expand it to git ref.
        If no ref given or isn't globbing pattern, returns it.

        Args:
            ref: git ref to be expanded if necessary

        Returns:
            same ref if is not globbing pattern, otherwise expanded
        """
        if not ref or not re.match(r".*[\[\?\*].*", ref):
            # regex matches any globbing pattern (`[`, `?` or `*` is used)
            logger.debug("No ref given or is not glob pattern")
            return ref

        tag = self.command_handler.run_command(
            get_current_version_command(
                ref.removeprefix("branches/"),
                refs="all" if ref.startswith("branches/") else "tags",
            ),
            return_output=True,
            cwd=self.local_project.working_dir,
        ).stdout.strip()
        logger.debug(f"Matching tag for {ref}: {tag}")

        return tag

    @staticmethod
    def _get_rpms_from_mock_output(output: str) -> list[str]:
        """
        Try to find the rpm files in the `mock` command output.

        Args:
            output: Output of the `mock` command.

        Returns:
            List of names of the RPM files.
        """
        logger.debug(f"The `mock` command output: {output}")
        reg = r"Results and/or logs in: (.*)(\s|$)"
        paths = re.findall(reg, output)

        rpms = [
            rpm.path
            for rpm in os.scandir(paths[0][0])
            if rpm.name.endswith(".rpm") and not rpm.name.endswith(".src.rpm")
        ]
        logger.debug(rpms)

        if not rpms:
            raise PackitRPMNotFoundException(
                "RPMs cannot be found, something is wrong.",
            )

        return rpms


class SRPMBuilder:
    def __init__(
        self,
        upstream: GitUpstream,
        srpm_path: Union[Path, str, None] = None,
        srpm_dir: Union[Path, str, None] = None,
        ref: Optional[str] = None,
    ) -> None:
        self.upstream = upstream
        self.srpm_path = srpm_path
        self.__ref = ref

        self._current_version = None
        self._upstream_ref = None

        if self.upstream.running_in_service():
            self.srpm_dir = Path(".")
            self.rpmbuild_dir = self.upstream.absolute_specfile_dir.relative_to(
                self.upstream.local_project.working_dir,
            )
        else:
            self.srpm_dir = Path(srpm_dir) if srpm_dir else Path.cwd()
            self.rpmbuild_dir = self.upstream.absolute_specfile_dir

    @property
    def current_version(self):
        if self._current_version is None:
            self._current_version = self.upstream.get_current_version()
        return self._current_version

    @property
    def upstream_ref(self):
        if self._upstream_ref is None:
            self._upstream_ref = self.upstream._expand_git_ref(
                self.__ref or self.upstream.package_config.upstream_ref,
            )
        return self._upstream_ref

    def _get_srpm_from_rpmbuild_output(self, output: str) -> str:
        """
        Try to find the SRPM file in the `rpmbuild -bs` command output.

        Args:
            output: Output of the `rpmbuild -bs` command.

        Returns:
            Name of the SRPM file.
        """
        logger.debug(f"The `rpmbuild` command output: {output}")
        # not doing 'Wrote: (.+)' since people can have different locales; hi Franto!
        reg = r": (\S+\.src\.rpm)"
        # also, we can't suffix this with '$' because rpmbuild can put additional content after
        # e.g. warnings when parsing the spec file
        try:
            the_srpm = re.findall(reg, output)[0]
        except IndexError as e:
            raise PackitSRPMNotFoundException(
                "SRPM cannot be found, something is wrong.",
            ) from e
        return the_srpm

    def get_build_command(self) -> tuple[list[str], str]:
        """
        Constructs `rpmbuild` command.

        Returns:
            Escaped `rpmbuild` command.
        """
        cmd = [
            "rpmbuild",
            "-bs",
            "--define",
            f"_sourcedir {self.rpmbuild_dir}",
            "--define",
            f"_srcdir {self.rpmbuild_dir}",
            "--define",
            f"_specdir {self.rpmbuild_dir}",
            "--define",
            f"_srcrpmdir {self.srpm_dir}",
            "--define",
            f"_topdir {self.rpmbuild_dir}",
            # we also need these 3 so that rpmbuild won't create them
            "--define",
            f"_builddir {self.rpmbuild_dir}",
            "--define",
            f"_rpmdir {self.rpmbuild_dir}",
            "--define",
            f"_buildrootdir {self.rpmbuild_dir}",
            self.upstream.package_config.specfile_path,
        ]
        escaped_command = " ".join(cmd)

        logger.debug(f"SRPM build command: {escaped_command}")
        return cmd, escaped_command

    def get_path(self, out: str) -> Path:
        """
        Get path to the SRPM file. In case it is necessary to move the built SRPM,
        move it.

        Args:
            out: Output of the `rpmbuild` command.

        Returns:
            Path to the SRPM.
        """
        built_srpm_path = self._get_srpm_from_rpmbuild_output(out)
        if self.srpm_path:
            shutil.move(built_srpm_path, self.srpm_path)
            return Path(self.srpm_path)

        if self.upstream.running_in_service():
            return self.upstream.local_project.working_dir / built_srpm_path
        return Path(built_srpm_path)

    def build(self) -> Path:
        cmd, escaped_command = self.get_build_command()

        present_srpms = set(self.srpm_dir.glob("*.src.rpm"))
        logger.debug(f"Present SRPMs: {present_srpms}")

        try:
            out = self.upstream.command_handler.run_command(
                cmd,
                return_output=True,
            ).stdout.strip()
        except PackitCommandFailedError as ex:
            logger.error(f"The `rpmbuild` command failed: {ex!r}")
            raise PackitFailedToCreateSRPMException(
                f"reason:\n{ex}\n"
                f"command:\n{escaped_command}\n"
                f"stdout:\n{ex.stdout_output}\n"
                f"stderr:\n{ex.stderr_output}",
            ) from ex
        except PackitException as ex:
            logger.error(f"The `rpmbuild` command failed: {ex!r}")
            raise PackitFailedToCreateSRPMException(
                f"The `rpmbuild` command failed:\n{ex}",
            ) from ex

        return self.get_path(out)

    def _prepare_upstream_using_source_git(
        self,
        update_release: bool,
        release_suffix: Optional[str],
    ) -> None:
        """
        Fetch the tarball and don't check out the upstream ref.
        """
        self.upstream.fetch_upstream_archive()
        self.upstream.create_patches_and_update_specfile(self.upstream_ref)

        ChangelogHelper(
            self.upstream,
            package_config=self.upstream.package_config,
        ).prepare_upstream_using_source_git(
            update_release,
            release_suffix,
        )

    def _fix_specfile_to_use_local_archive(
        self,
        archive: str,
        update_release: bool,
        release_suffix: Optional[str],
    ) -> None:
        """
        Update specfile to use the archive with the right version.

        Args:
            archive: Path to the archive.
            update_release: Should Release be updated?
            release_suffix: Append this suffix to the %release.
        """
        current_commit = self.upstream.local_project.commit_hexsha
        # the logic behind the naming:
        # * PACKIT - our namespace
        # * PACKIT_PROJECT - info about the project which we obtained
        # * PACKIT_RPMSPEC - data for the project's specfile assembled by us
        #                  - RPMSPEC is more descriptive than just SPEC
        env = {
            "PACKIT_PROJECT_VERSION": self.current_version,
            # Spec file %release field which packit sets by default
            "PACKIT_RPMSPEC_RELEASE": self.upstream.get_spec_release(release_suffix),
            "PACKIT_PROJECT_COMMIT": current_commit,
            "PACKIT_PROJECT_ARCHIVE": archive,
            "PACKIT_PROJECT_BRANCH": sanitize_version(
                self.upstream.local_project.ref,
            ),
        }

        # in case we are given template as a release suffix
        if release_suffix and reduce(
            lambda has_macro, macro: has_macro or (macro in release_suffix),
            env.keys(),
            False,
        ):
            # The release_suffix contains macros to be expanded
            # do not use it to format the PACKIT_RPMSPEC_RELEASE!
            # Otherwise, you will obtain something like
            # 0.{PACKIT_RPMSPEC_RELEASE} as result.
            # In this case PACKIT_RPMSPEC_RELEASE should be expanded
            # like when no release_suffix is given: so use "" instead.
            env["PACKIT_RPMSPEC_RELEASE"] = self.upstream.get_spec_release("")
            new_release = release_suffix.format(**env)
        else:
            new_release = self.upstream.get_spec_release(release_suffix)
        env = env | self.upstream.package_config.get_package_names_as_env()

        if self.upstream.actions_handler.with_action(
            action=ActionName.fix_spec,
            env=env,
        ):
            self.upstream.fix_spec(
                archive=archive,
                version=self.current_version,
                commit=current_commit,
                update_release=update_release,
                release=new_release,
            )
        self.upstream.specfile.reload()  # the specfile could have been changed by the action

    def prepare(
        self,
        update_release: bool,
        release_suffix: Optional[str] = None,
        create_symlinks: Optional[bool] = True,
    ):
        if self.upstream_ref:
            self._prepare_upstream_using_source_git(update_release, release_suffix)
        else:
            created_archive = self.upstream.create_archive(
                version=self.current_version,
                create_symlink=create_symlinks,
            )
            self._fix_specfile_to_use_local_archive(
                archive=created_archive,
                update_release=update_release,
                release_suffix=release_suffix,
            )

        # https://github.com/packit/packit-service/issues/314
        if Path(self.upstream.local_project.working_dir).joinpath("sources").exists():
            logger.warning('The upstream repo contains "sources" file or a directory.')
            logger.warning(
                "We are unable to download remote sources from spec-file "
                "because the file contains links to archives in Fedora downstream.",
            )
            logger.warning("Therefore skipping downloading of remote sources.")
        else:
            self.upstream.download_remote_sources()


class Archive:
    def __init__(self, upstream: GitUpstream, version: Optional[str] = None) -> None:
        """
        Creates an instance of `Archive`.

        Args:
            upstream: Instance of Upstream class.
            version: Version of the archive.

                Defaults to `None`.
        """
        self.upstream = upstream
        self._version = version

    @property
    def version(self):
        """
        Version of the archive. If not given through constructor, initialized with
        `get_current_version` from `Upstream` class.
        """
        if not self._version:
            self._version = self.upstream.get_current_version()
        return self._version

    def create(self, create_symlink: Optional[bool] = True) -> str:
        """
        Create archive from the content of the upstream repository, only committed
        changes are present in the archive.

        Uses `git archive` by default, unless `create_archive` action is defined.

        Args:
            create_symlink: whether symlink to archive should be created when the
                created archive is outside the specfile dir or the archive should be copied

        Returns:
            Name of the archive.
        """
        package_name = (
            self.upstream.package_config.upstream_package_name
            or self.upstream.package_config.downstream_package_name
        )
        dir_name = f"{package_name}-{self.version}"
        logger.debug(f"Name + version = {dir_name}")

        env = {
            "PACKIT_PROJECT_VERSION": self.version,
            "PACKIT_PROJECT_NAME_VERSION": dir_name,
        }
        env = env | self.upstream.package_config.get_package_names_as_env()
        if self.upstream.actions_handler.has_action(action=ActionName.create_archive):
            outputs = self.upstream.actions_handler.get_output_from_action(
                action=ActionName.create_archive,
                env=env,
            )
            if not outputs:
                raise PackitException("No output from create-archive action.")

            archive_path = self._get_archive_path_from_output(outputs)
            if not archive_path:
                raise PackitException(
                    "The create-archive action did not output a path to the generated archive. "
                    "Please make sure that you have valid path in the single line of the output.",
                )
            self._handle_archive_outside_specdir_if_needed(archive_path, create_symlink)
            return archive_path.name

        return self._create_archive_using_default_way(dir_name, env)

    def _get_archive_path_from_line(self, line: str) -> Optional[Path]:
        """
        Get path to the created archive from one line of the output.

        Args:
            line: Line of output produced while creating the archive.

        Returns:
            Path to the archive as string, `None` if cannot be parsed.
        """
        try:
            archive_path = Path(line.strip())

            if not archive_path.is_absolute():
                archive_path = self.upstream._local_project.working_dir / archive_path

            if archive_path.is_file():
                archive_path_absolute = archive_path.absolute()
                logger.info("Created archive:")
                logger.info(f"\tparsed   path: {archive_path}")
                logger.info(f"\tabsolute path: {archive_path_absolute}")
                return archive_path_absolute
        except OSError as ex:
            # File too long
            if ex.errno == 36:
                logger.error(
                    "Skipping long output command output while getting archive name.",
                )
                return None
            raise ex

        return None

    def _get_archive_path_from_output(self, outputs: list[str]) -> Optional[Path]:
        """
        Parse the archive name from the output in the reverse order.
        Check if the line is a path and if it exists.

        Args:
            outputs: Outputs produced by creating the archive.

        Returns:
            Absolute path to the archive if found, `None` otherwise.
        """
        for output in reversed(outputs):
            for line in reversed(output.splitlines()):
                archive_path = self._get_archive_path_from_line(line)
                if archive_path:
                    return archive_path
        return None

    def _handle_archive_outside_specdir_if_needed(
        self,
        archive_path: Path,
        create_symlink: Optional[bool] = True,
    ) -> None:
        """
        Create a relative symlink to the archive from in the specfile directory
        or copy the archive to the specfile directory if necessary.

        Args:
            archive_path: Absolute path to the archive from the specfile dir.
            create_symlink: Whether a symlink should be created.
        """
        absolute_specfile_dir = self.upstream.absolute_specfile_dir

        if archive_path.parent.absolute() != absolute_specfile_dir:
            archive_in_spec_dir = absolute_specfile_dir / archive_path.name

            if create_symlink:
                # [PurePath.relative_to()](https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.relative_to)
                # requires self to be the subpath of the argument, but
                # [os.path.relpath()](https://docs.python.org/3/library/os.path.html#os.path.relpath)
                # does not.
                relative_archive_path = os.path.relpath(
                    archive_path,
                    absolute_specfile_dir,
                )

                logger.info(
                    "Linking to the specfile directory:"
                    f" {archive_in_spec_dir} -> {relative_archive_path}"
                    f" (given path to archive: {archive_path})",
                )
                archive_in_spec_dir.symlink_to(relative_archive_path)

            else:
                logger.info(
                    "Copying the archive to the specfile directory: "
                    f"{archive_path} -> {archive_in_spec_dir}",
                )
                shutil.copy2(archive_path, archive_in_spec_dir)

    def _create_archive_using_default_way(
        self,
        dir_name: str,
        env: dict[str, str],
    ) -> str:
        """
        Create an archive using `git archive`.
        Archive will be placed in the `specfile_directory`.

        Args:
            dir_name: Name of the directory from which the archive is created.
            env: Environment variables passed to the action.

        Returns:
            Name of the archive as a string.
        """
        archive_name = f"{dir_name}{DEFAULT_ARCHIVE_EXT}"
        relative_archive_path = (
            self.upstream.absolute_specfile_dir / archive_name
        ).relative_to(self.upstream.local_project.working_dir)
        archive_cmd = [
            "git",
            "archive",
            "--output",
            str(relative_archive_path),
            "--prefix",
            f"{dir_name}/",
            "HEAD",
        ]
        self.upstream.command_handler.run_command(
            archive_cmd,
            return_output=True,
            env=env,
        )
        return archive_name

    def get_archive_root_dir(self, archive: str) -> Optional[str]:
        """
        Get archive's root directory.

        It uses 2 techniques:
        1. tries to extract it directly from archive.
        2. will generate name based on `archive_root_dir_template`

        Currently supported archives:
        * tar including compression (for details check python tarfile module doc)

        Args:
            archive: Name of the archive.

        Returns:
            Archive's top-level directory or `None`.

        Raises:
            PackitException: If failed, i.e. all methods used for deduction returned
                `None`.
        """

        archive_root_dir = None

        logger.debug("Trying to extract archive_root_dir from known archives")
        if tarfile.is_tarfile(f"{self.upstream.absolute_specfile_dir}/{archive}"):
            logger.debug(f"Archive {archive} is tar.")
            archive_root_dir = self.get_archive_root_dir_from_tar(archive)
        else:
            logger.debug(f"Archive {archive} is not tar.")

        if archive_root_dir is None:
            logger.debug(
                "Using archive_root_dir_template config option. If not set it defaults to "
                "{{upstream_pkg_name}}-{{version}}. Check "
                "https://packit.dev/docs/configuration/#archive_root_dir_template "
                "for more details.",
            )
            archive_root_dir = self.get_archive_root_dir_from_template()

        return archive_root_dir

    def get_archive_root_dir_from_tar(self, archive: str) -> Optional[str]:
        """
        Returns tar archive's top-level directory, if there is exactly one.

        Args:
            archive: Name of the tar archive.

        Returns:
            Archive's top level directory if there is exactly one, `None` otherwise.
        """

        tar = tarfile.open(f"{self.upstream.absolute_specfile_dir}/{archive}")
        root_dirs = set()
        for tar_item in tar.getmembers():
            if tar_item.isdir() and "/" not in tar_item.name:
                root_dirs.add(tar_item.name)
            # required for archives where top-level dir was added using tar --transform
            # option - in that case, tar archive will not contain dir related entry
            if tar_item.isfile() and "/" in tar_item.name:
                root_dirs.add(tar_item.name.split("/")[0])

        root_dirs_count = len(root_dirs)
        archive_root_items_count = len(
            {i.name for i in tar.getmembers() if "/" not in i.name},
        )

        if root_dirs_count == 1:
            root_dir = root_dirs.pop()
            logger.debug(f"Directory {root_dir} found in archive {archive}")
            return root_dir

        if root_dirs_count == 0:
            logger.warning(f"No directory found in archive {archive}.")
        elif root_dirs_count > 1:
            logger.warning(
                f"Archive {archive} contains multiple directories on the top level: "
                f"the common practice in the industry is to have only one in the "
                f'following format: "PACKAGE-VERSION"',
            )
        elif archive_root_items_count > 1:
            logger.warning(
                f"Archive f{archive} contains multiple root items. It can be "
                f"intentional or can signal incorrect archive structure.",
            )

        return None

    def get_archive_root_dir_from_template(self) -> Optional[str]:
        """
        Generates archive's root directory based on the `archive_root_dir_template`.
        `archive_root_dir_template`'s default value is `{upstream_pkg_name}-{version}`

        Returns:
            Archive's root directory name based on the template.
        """
        template = self.upstream.package_config.archive_root_dir_template
        logger.debug(
            f"archive_root_dir_template is set or defaults to if not set to: {template}",
        )
        archive_root_dir = template.replace(
            "{upstream_pkg_name}",
            self.upstream.package_config.upstream_package_name,
        ).replace("{version}", self.version)
        not_replaced = re.findall("{.*?}", archive_root_dir)
        if not_replaced:
            logger.warning(
                f"Probably not all archive_root_dir_template tags were "
                f"replaced: {' ,'.join(not_replaced)}",
            )
        return archive_root_dir



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""The 'build' subcommand for Packit"""

import logging

import click

from packit.cli.builds.copr_build import copr
from packit.cli.builds.in_image_builder import in_image_builder
from packit.cli.builds.koji_build import koji
from packit.cli.builds.local_build import local
from packit.cli.builds.mock_build import mock

logger = logging.getLogger(__name__)


@click.group("build")
@click.option(
    "--srpm",
    help="Build the SRPM from FILE instead of implicit SRPM build.",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
)
@click.pass_context
def build(ctx, srpm):
    """Subcommand to collect build related functionality"""
    ctx.obj.srpm_path = srpm


build.add_command(copr)
build.add_command(koji)
build.add_command(local)
build.add_command(mock)
build.add_command(in_image_builder)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
from datetime import timedelta
from typing import Union

from cachetools.func import ttl_cache

from packit.constants import FAST_FORWARD_MERGE_INTO_KEY
from packit.exceptions import PackitException
from packit.utils.bodhi import get_bodhi_client
from packit.utils.commands import run_command
from packit.utils.decorators import fallback_return_value
from packit.utils.monitoring import Pushgateway

ALIASES: dict[str, list[str]] = {
    "fedora-all": ["fedora-39", "fedora-40", "fedora-41", "fedora-rawhide"],
    "fedora-stable": ["fedora-39", "fedora-40"],
    "fedora-development": ["fedora-41", "fedora-rawhide"],
    "fedora-latest": ["fedora-41"],
    "fedora-latest-stable": ["fedora-40"],
    "fedora-branched": ["fedora-39", "fedora-40", "fedora-41"],
    "epel-all": ["epel-8", "epel-9", "epel-10.0"],
}

ARCHITECTURE_LIST: list[str] = [
    "aarch64",
    "armhfp",
    "i386",
    "ppc64le",
    "s390x",
    "x86_64",
]

DEPRECATED_TARGET_MAP = {"centos-stream": "centos-stream-8"}
DEPRECATED_TARGET_MAP = {
    f"{k}-{x}": f"{v}-{x}"
    for k, v in DEPRECATED_TARGET_MAP.items()
    for x in ARCHITECTURE_LIST
}

DEFAULT_VERSION = "fedora-stable"

logger = logging.getLogger(__name__)


pushgateway = Pushgateway()


def get_versions(*name: str, default=DEFAULT_VERSION) -> set[str]:
    """
    Expand the aliases to the name(s).

    :param name: name(s) of the system and version (e.g. "fedora-30" or "fedora-stable")
    :param default: used if no positional argument was given
    :return: set of string containing system name and version
    """
    if not (default or name):
        return set()

    names = list(name) or [default]
    versions: set[str] = set()
    for one_name in names:
        versions.update(get_aliases().get(one_name, [one_name]))
    return versions


def get_build_targets(*name: str, default: str = DEFAULT_VERSION) -> set[str]:
    """
    Expand the aliases to the name(s) and transfer to the build targets.

    :param name: name(s) of the system and version (e.g. "fedora-30" or "fedora-stable")
            or target name (e.g. "fedora-30-x86_64" or "fedora-stable-x86_64")
    :param default: used if no positional argument was given
    :return: set of build targets
    """
    if not (default or name):
        return set()

    names = list(name) or [default]
    possible_sys_and_versions: set[str] = set()
    for one_name in names:
        name_split = one_name.rsplit("-", maxsplit=2)
        l_name_split = len(name_split)

        if l_name_split < 2:  # only one part
            # => cannot guess anything other than rawhide
            if "rawhide" in one_name:
                sys_name, version, architecture = "fedora", "rawhide", "x86_64"

            else:
                err_msg = (
                    f"Cannot get build target from '{one_name}'"
                    f", packit understands values like these: '{list(get_aliases().keys())}'."
                )
                raise PackitException(err_msg.format(one_name=one_name))

        elif l_name_split == 2:  # "name-version"
            sys_name, version = name_split
            architecture = "x86_64"  # use the x86_64 as a default

        else:  # "name-version-architecture"
            sys_name, version, architecture = name_split
            if architecture not in ARCHITECTURE_LIST:
                # we don't know the architecture => probably wrongly parsed
                # (e.g. "opensuse-leap-15.0")
                sys_name, version, architecture = (
                    f"{sys_name}-{version}",
                    architecture,
                    "x86_64",
                )

        possible_sys_and_versions.update(
            {
                f"{sys_and_version}-{architecture}"
                for sys_and_version in get_versions(f"{sys_name}-{version}")
            },
        )
    return {
        DEPRECATED_TARGET_MAP.get(target, target)
        for target in possible_sys_and_versions
    }


def get_branches(
    *name: str,
    default: str = DEFAULT_VERSION,
    default_dg_branch: str = "main",
    with_aliases: bool = False,
) -> set[str]:
    """
    Expand the aliases to the name(s) and transfer to the dist-git branch name.

    Args:
        name: Name(s) of the system and version (e.g. "fedora-stable"
            or "fedora-30") or branch name (e.g. "f30" or "epel8").
        default: Used if no positional argument was given.
        default_dg_branch: Default branch of dist-git repository.
        with_aliases: If set to `True`, returns branches including aliases.
            Can be used for reacting to webhooks where push can be done against
            either the branch itself or its alias.

            Defaults to `False`.

    Returns:
        Set of dist-git branch names.
    """
    if not (default or name):
        return set()

    names = list(name) or [default]
    branches = set()

    for sys_and_version in get_versions(*names):
        if "rawhide" in sys_and_version:
            branches.add(default_dg_branch)
            if with_aliases:
                branches.add("rawhide")
        elif sys_and_version.startswith("fedora"):
            sys, version = sys_and_version.rsplit("-", maxsplit=1)
            branches.add(f"f{version}")
        elif sys_and_version.startswith("epel"):
            split = sys_and_version.rsplit("-", maxsplit=1)
            if len(split) < 2:
                branches.add(sys_and_version)
                continue
            sys, version = sys_and_version.rsplit("-", maxsplit=1)
            if version.isnumeric() and int(version) <= 6:
                branches.add(f"el{version}")
            else:
                branches.add(f"epel{version}")
        else:
            # We don't know, let's leave the original name.
            branches.add(sys_and_version)

    return branches


def get_fast_forward_merge_branches_for(
    dist_git_branches: Union[list, dict, set, None],
    source_branch: str,
    default: str = DEFAULT_VERSION,
    default_dg_branch: str = "main",
    with_aliases: bool = False,
) -> set[str]:
    """
    Returns a list of target branches that can be fast forwarded merging
    the specified source_branch
    The keys can be aliases, expand them into a temporary structure
    to be sure that source_branch can be found in the first dictionary.
    In the inner loop of the downstream sync the expanded aliases are used.

    The branches should be specified as in the following example
        {"rawhide": {"fast_forward_merge_into": ["f40", "f39"]},
         "epel9": {}
        }

    dist_git_branches: dist_git_branches config key; can be a list or dict
    source_branch: source branch with commits to be merged
    default: the same as get_branches default parameter
    default_dg_branch: the same as get_branches default_dg_branch parameter
    with_aliases: the same as get_branches with_aliases parameter
    """
    if not isinstance(dist_git_branches, dict):
        return set()

    expanded_dist_git_branches = {}
    for key, value in dist_git_branches.items():
        expanded_keys = get_branches(
            *[key],
            default=default,
            default_dg_branch=default_dg_branch,
            with_aliases=True,
        )
        expanded_dist_git_branches.update(
            {expanded_key: value for expanded_key in expanded_keys},
        )

    if source_branch not in expanded_dist_git_branches:
        return set()

    if (source_branch_dict := expanded_dist_git_branches[source_branch]) and isinstance(
        source_branch_dict[FAST_FORWARD_MERGE_INTO_KEY],
        list,
    ):
        return get_branches(
            *source_branch_dict[FAST_FORWARD_MERGE_INTO_KEY],
            default=default,
            default_dg_branch=default_dg_branch,
            with_aliases=with_aliases,
        )

    return set()


def get_koji_targets(*name: str, default=DEFAULT_VERSION) -> set[str]:
    if not (default or name):
        return set()

    names = list(name) or [default]
    targets = set()

    for sys_and_version in get_versions(*names):
        if sys_and_version == "fedora-rawhide":
            targets.add("rawhide")
        elif sys_and_version.startswith("fedora"):
            sys, version = sys_and_version.rsplit("-", maxsplit=1)
            targets.add(f"f{version}")
        elif sys_and_version.startswith("el") and sys_and_version[2:].isnumeric():
            targets.add(f"epel{sys_and_version[2:]}")
        elif sys_and_version.startswith("epel"):
            split = sys_and_version.rsplit("-", maxsplit=1)
            if len(split) == 2:
                sys, version = split
                if version.isnumeric():
                    targets.add(f"epel{version}")
                    continue
            targets.add(sys_and_version)
        else:
            # We don't know, let's leave the original name.
            targets.add(sys_and_version)

    return targets


def get_all_koji_targets() -> list[str]:
    return run_command(["koji", "list-targets", "--quiet"], output=True).stdout.split()


def update_metrics():
    if hasattr(pushgateway, "aliases_fallback_used"):
        pushgateway.aliases_fallback_used.inc()
    pushgateway.push()


@ttl_cache(maxsize=1, ttl=timedelta(hours=12).seconds)
@fallback_return_value(ALIASES, callback=update_metrics)
def get_aliases() -> dict[str, list[str]]:
    """
    Function to automatically determine fedora-* and epel-* aliases.
    Current data are fetched via bodhi client, with default base url
    `https://bodhi.fedoraproject.org/'.

    Returns:
        Dictionary containing aliases.
    """
    bodhi_client = get_bodhi_client()
    releases = []
    page = pages = 1
    while page <= pages:
        results = bodhi_client.get_releases(exclude_archived=True, page=page)
        releases.extend(results.releases)
        page += 1
        pages = results.pages
    current_fedora_releases, pending_fedora_releases, epel_releases = [], [], []

    for release in filter(
        lambda r: r.state in ["current", "pending", "frozen"],
        releases,
    ):
        if release.id_prefix == "FEDORA" and release.name != "ELN":
            name = release.long_name.lower().replace(" ", "-")
            if release.state == "current":
                current_fedora_releases.append(name)
            else:
                pending_fedora_releases.append(name)
        elif release.id_prefix == "FEDORA-EPEL":
            name = release.name.lower()
            epel_releases.append(name)

    current_fedora_releases.sort(key=lambda x: int(x.rsplit("-")[-1]))
    pending_fedora_releases.sort(key=lambda x: int(x.rsplit("-")[-1]))
    # The Fedora with the highest version is "rawhide", but
    # Bodhi always uses release names, and has no concept of "rawhide".
    pending_fedora_releases[-1] = "fedora-rawhide"

    #  fedora-34   fedora-35   [ fedora-36 ]   fedora-rawhide
    #  current     current  [ current/pending ]  pending
    #
    # all: everything
    # stable: everything marked as "current"
    # development: everything marked as "pending"
    # latest: the latest with a version number
    # latest-stable: the last "current"
    # branched: all with a version number
    return {
        "fedora-all": current_fedora_releases + pending_fedora_releases,
        "fedora-stable": current_fedora_releases,
        "fedora-development": pending_fedora_releases,
        "fedora-latest": pending_fedora_releases[-2:-1] or current_fedora_releases[-1:],
        "fedora-latest-stable": current_fedora_releases[-1:],
        "fedora-branched": current_fedora_releases + pending_fedora_releases[:-1],
        "epel-all": epel_releases,
    }



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
from os import getcwd

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api, iterate_packages
from packit.config import get_context_settings, pass_config
from packit.config.aliases import get_branches, get_koji_targets
from packit.constants import (
    PACKAGE_LONG_OPTION,
    PACKAGE_OPTION_HELP,
    PACKAGE_SHORT_OPTION,
)
from packit.exceptions import (
    PackitCommandFailedError,
    PackitConfigException,
    ensure_str,
)
from packit.utils.changelog_helper import ChangelogHelper

logger = logging.getLogger(__name__)


@click.command("in-koji", context_settings=get_context_settings())
@click.option(
    "--dist-git-branch",
    help="Comma separated list of target branches in dist-git to release into. "
    "(defaults to repo's default branch)",
)
@click.option(
    "--dist-git-path",
    help="Path to dist-git repo to work in. "
    "Otherwise clone the repo in a temporary directory.",
)
@click.option(
    "--from-upstream",
    help="Build the project in koji directly from the upstream repository",
    is_flag=True,
    default=False,
)
@click.option(
    "--koji-target",
    help="Koji target to build inside (see `koji list-targets`).",
)
@click.option(
    "--scratch",
    is_flag=True,
    default=False,
    help="Submit a scratch koji build",
)
@click.option("--wait/--no-wait", default=True, help="Wait for the build to finish")
@click.option(
    "--release-suffix",
    default=None,
    type=click.STRING,
    help="Specifies release suffix. Allows to override default generated:"
    "{current_time}.{sanitized_current_branch}{git_desc_suffix}",
)
@click.option(
    "--default-release-suffix",
    default=False,
    is_flag=True,
    help=(
        "Allows to use default, packit-generated, release suffix when some "
        "release_suffix is specified in the configuration."
    ),
)
@click.option(
    PACKAGE_SHORT_OPTION,
    PACKAGE_LONG_OPTION,
    multiple=True,
    help=PACKAGE_OPTION_HELP.format(action="build"),
)
@click.argument("path_or_url", type=LocalProjectParameter(), default=getcwd())
@pass_config
@cover_packit_exception
@iterate_packages
def koji(
    config,
    dist_git_path,
    dist_git_branch,
    from_upstream,
    koji_target,
    scratch,
    wait,
    release_suffix,
    default_release_suffix,
    package_config,
    path_or_url,
):
    """
    Build selected upstream project in Fedora.

    By default, packit checks out the respective dist-git repository and performs
    `fedpkg build` for the selected branch. With `--from-upstream`, packit creates a SRPM
    out of the current checkout and sends it to koji.

    PATH_OR_URL argument is a local path or a URL to the upstream git repository,
    it defaults to the current working directory
    """
    api = get_packit_api(
        config=config,
        package_config=package_config,
        dist_git_path=dist_git_path,
        local_project=path_or_url,
    )
    release_suffix = ChangelogHelper.resolve_release_suffix(
        api.package_config,
        release_suffix,
        default_release_suffix,
    )

    default_dg_branch = api.dg.local_project.git_project.default_branch
    dist_git_branch = dist_git_branch or default_dg_branch
    branches_to_build = get_branches(
        *dist_git_branch.split(","),
        default_dg_branch=default_dg_branch,
    )
    package = (
        package_config.downstream_package_name
        if package_config.downstream_package_name
        else package_config.upstream_package_name
    )
    click.echo(
        f"Building from the following branches: {', '.join(branches_to_build)}, "
        f"for package {package}",
    )

    targets_to_build = {None} if koji_target is None else get_koji_targets(koji_target)

    if len(targets_to_build) > 1 and len(branches_to_build) > 1:
        raise PackitConfigException(
            "Parameters --dist-git-branch and --koji-target cannot have "
            "multiple values at the same time.",
        )

    for target in targets_to_build:
        for branch in branches_to_build:
            try:
                out = api.build(
                    dist_git_branch=branch,
                    scratch=scratch,
                    nowait=not wait,
                    koji_target=target,
                    from_upstream=from_upstream,
                    release_suffix=release_suffix,
                    srpm_path=config.srpm_path,
                )
            except PackitCommandFailedError as ex:  # noqa: PERF203
                logs_stdout = "\n>>> ".join(ex.stdout_output.strip().split("\n"))
                logs_stderr = "\n!!! ".join(ex.stderr_output.strip().split("\n"))
                click.echo(
                    f"Build for branch '{branch}' failed. \n"
                    f">>> {logs_stdout}\n"
                    f"!!! {logs_stderr}\n",
                    err=True,
                )
            else:
                if out:
                    print(ensure_str(out))
